
# Can be run with python -m pie.scripts.train
import time
import os
from datetime import datetime
import logging

import pie
from pie.settings import settings_from_file
from pie.trainer import Trainer
from pie import initialization
from pie.data import Dataset, Reader, MultiLabelEncoder
from pie.models import SimpleModel, get_pretrained_embeddings
from pie import optimize

# set seeds
import random
import numpy
import torch


def get_targets(settings):
    return [task['name'] for task in settings.tasks if task.get('target')]


def get_fname_infix(settings):
    # fname
    fname = os.path.join(settings.modelpath, settings.modelname)
    timestamp = datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
    infix = '+'.join(get_targets(settings)) + '-' + timestamp
    return fname, infix


def run(settings):
    now = datetime.now()

    # set seed
    seed = now.hour * 10000 + now.minute * 100 + now.second
    print("Using seed:", seed)
    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)

    if settings.verbose:
        logging.basicConfig(level=logging.INFO)

    # datasets
    reader = Reader(settings, settings.input_path)
    tasks = reader.check_tasks(expected=None)
    if settings.verbose:
        print("::: Available tasks :::")
        print()
        for task in tasks:
            print("- {}".format(task))
        print()

    # label encoder
    label_encoder = MultiLabelEncoder.from_settings(settings, tasks=tasks)
    if settings.verbose:
        print("::: Fitting data :::")
        print()
    label_encoder.fit_reader(reader)

    if settings.verbose:
        print()
        print("::: Vocabulary :::")
        print()
        types = '{}/{}={:.2f}'.format(*label_encoder.word.get_type_stats())
        tokens = '{}/{}={:.2f}'.format(*label_encoder.word.get_token_stats())
        print("- {:<15} types={:<10} tokens={:<10}".format("word", types, tokens))
        types = '{}/{}={:.2f}'.format(*label_encoder.char.get_type_stats())
        tokens = '{}/{}={:.2f}'.format(*label_encoder.char.get_token_stats())
        print("- {:<15} types={:<10} tokens={:<10}".format("char", types, tokens))
        print()
        print("::: Tasks :::")
        print()
        for task, le in label_encoder.tasks.items():
            print("- {:<15} target={:<6} level={:<6} vocab={:<6}"
                  .format(task, le.target, le.level, len(le)))
        print()

    trainset = Dataset(settings, reader, label_encoder)

    devset = None
    if settings.dev_path:
        devset = Dataset(settings, Reader(settings, settings.dev_path), label_encoder)
    else:
        logging.warning("No devset: cannot monitor/optimize training")

    # model
    model = SimpleModel(
        label_encoder, settings.tasks,
        settings.wemb_dim, settings.cemb_dim, settings.hidden_size,
        settings.num_layers, cell=settings.cell,
        # dropout
        dropout=settings.dropout, word_dropout=settings.word_dropout,
        # word embeddings
        merge_type=settings.merge_type, cemb_type=settings.cemb_type,
        cemb_layers=settings.cemb_layers, custom_cemb_cell=settings.custom_cemb_cell,
        # lm joint loss
        include_lm=settings.include_lm, lm_shared_softmax=settings.lm_shared_softmax,
        # decoder
        scorer=settings.scorer, linear_layers=settings.linear_layers)

    # pretrain(/load pretrained) embeddings
    if model.wemb is not None:
        if settings.pretrain_embeddings:
            print("Pretraining word embeddings")
            wemb_reader = Reader(
                settings, settings.input_path, settings.dev_path, settings.test_path)
            weight = get_pretrained_embeddings(
                wemb_reader, label_encoder, size=settings.wemb_dim,
                window=5, negative=5, min_count=1)
            model.wemb.weight.data = torch.tensor(weight, dtype=torch.float32)

        elif settings.load_pretrained_embeddings:
            print("Loading pretrained embeddings")
            if not os.path.isfile(settings.load_pretrained_embeddings):
                print("Couldn't find pretrained eembeddings in: {}".format(
                    settings.load_pretrained_embeddings))
            initialization.init_pretrained_embeddings(
                settings.load_pretrained_embeddings, label_encoder.word, model.wemb)

    # load pretrained weights
    if settings.load_pretrained_encoder:
        model.init_from_encoder(pie.Encoder.load(settings.load_pretrained_encoder))

    # freeze embeddings
    if settings.freeze_embeddings:
        model.wemb.weight.requires_grad = False

    model.to(settings.device)

    print("::: Model :::")
    print()
    print(model)
    print()
    print("::: Model parameters :::")
    print()
    trainable = sum(p.nelement() for p in model.parameters() if p.requires_grad)
    total = sum(p.nelement() for p in model.parameters())
    print("{}/{} trainable/total".format(trainable, total))
    print()

    # training
    print("Starting training")

    running_time = time.time()
    trainer = Trainer(settings, model, trainset, reader.get_nsents())
    scores = None
    try:
        scores = trainer.train_epochs(settings.epochs, devset=devset)
    except KeyboardInterrupt:
        print("Stopping training")
    finally:
        model.eval()
    running_time = time.time() - running_time

    if settings.test_path:
        print("Evaluating model on test set")
        testset = Dataset(settings, Reader(settings, settings.test_path), label_encoder)
        for task in model.evaluate(testset, trainset).values():
            task.print_summary()

    # save model
    fpath, infix = get_fname_infix(settings)
    if not settings.run_test:
        fpath = model.save(fpath, infix=infix, settings=settings)
        print("Saved best model to: [{}]".format(fpath))

    if devset is not None and not settings.run_test:
        scorers = model.evaluate(devset, trainset)
        scores = []
        for task in sorted(scorers):
            scorer = scorers[task]
            result = scorer.get_scores()
            for acc in result:
                scores.append('{}-{}:{:.6f}'.format(
                    acc, task, result[acc]['accuracy']))
                scores.append('{}-{}-support:{}'.format(
                    acc, task, result[acc]['support']))
        path = '{}.results.{}.csv'.format(
            settings.modelname, '-'.join(get_targets(settings)))
        with open(path, 'a') as f:
            line = [infix, str(seed), str(running_time)]
            line += scores
            f.write('{}\n'.format('\t'.join(line)))

    print("Bye!")


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('config_path', nargs='?', default='config.json')
    parser.add_argument('--opt_path', help='Path to optimization file (see opt.json)')
    parser.add_argument('--n_iter', type=int, default=20)
    args = parser.parse_args()

    settings = settings_from_file(args.config_path)

    if args.opt_path:
        opt = optimize.read_opt(args.opt_path)
        optimize.run_optimize(run, settings, opt, args.n_iter)
    else:
        run(settings)




import os
import uuid
import logging
import time
import collections
import random
import tempfile


import tqdm

import torch
from torch import optim
from torch.nn.utils import clip_grad_norm_

logging.basicConfig(format='%(asctime)s : %(message)s', level=logging.INFO)


def get_batch_task(tasks, **kwargs):
    tnames, target = [], None
    for task in tasks:
        tnames.append(task['name'])
        if task.get('target'):
            target = task['name']
    return sample_task(target, tnames, **kwargs)


def sample_task(target, tasks, factor=2):
    # sample target task factor times as many as any other task
    aux = (1 / (len(tasks) - 1)) / factor if len(tasks) > 1 else 0
    trg = 1 - (aux * (len(tasks) - 1))
    weights = [aux if task != target else trg for task in tasks]
    return random.choices(tasks, weights)[0]


def get_target_task(settings):
    for task in settings.tasks:
        if task.get('target'):
            return task['name']
    raise ValueError("No target task?")


class EarlyStopException(Exception):
    def __init__(self, task, loss, state_dict):
        self.task = task
        self.loss = loss
        self.best_state_dict = state_dict


class TaskScheduler(object):
    """
    Track scores
    """
    def __init__(self, settings):
        tasks = {}
        # preprocess tasks
        for task in settings.tasks:
            # add schedule and target
            tasks[task['name']] = task.get('schedule', {})
            tasks[task['name']]['target'] = task.get('target', False)
            # add task data for lm loss
            if settings.include_lm:
                tasks['lm_fwd'] = dict(settings.lm_schedule)
                tasks['lm_bwd'] = dict(settings.lm_schedule)

        for task, tdata in tasks.items():
            # set step counter
            tdata['steps'] = 0
            # set default task mode
            tdata['mode'] = tdata.get('mode', 'max')
            # set initial weight
            tdata['weight'] = tdata.get('weight', 1.0)
            # set initial best
            tdata['best'] = -float('inf') if tdata['mode'] == 'max' else float('inf')

        self.tasks = tasks

        # task schedule
        self.patience = settings.patience
        self.factor = settings.factor
        self.threshold = settings.threshold
        self.min_weight = settings.min_weight
        self.fid = os.path.join(tempfile.gettempdir(), str(uuid.uuid1()))

    def __repr__(self):
        # task scheduler
        output = (
            '<TaskScheduler patience="{}" factor="{}" ' +
            'threshold="{}" min_weight="{}">').format(
                self.patience, self.factor, self.threshold, self.min_weight)

        for task, values in self.tasks.items():
            output += '\n    <Task name="{}" '.format(task)
            output += ' '.join(
                '{}="{}"'.format(key, val) for key, val in values.items())
            output += '/>'
        output += '\n</TaskScheduler>'

        return output

    def is_best(self, task, value):
        threshold = self.tasks[task].get('threshold', self.threshold)
        mode = self.tasks[task]['mode']
        if mode.lower() == 'max':
            return value > (self.tasks[task]['best'] + threshold)
        elif mode.lower() == 'min':
            return value < (self.tasks[task]['best'] - threshold)
        else:
            raise ValueError("Wrong mode value [{}] for task: {}".format(mode, task))

    def step(self, scores, model):
        """
        Advance schedule step based on dev scores
        """
        for task, score in scores.items():
            if task not in self.tasks:
                # ignore
                continue

            is_target = self.tasks[task].get('target', False)

            # check if we improve
            if self.is_best(task, score):
                self.tasks[task]['best'] = score
                self.tasks[task]['steps'] = 0
                if is_target:
                    # serialize model params
                    torch.save(model.state_dict(), self.fid)
            else:
                self.tasks[task]['steps'] += 1

            # check if we need to stop globally or downweight a task loss
            patience = self.tasks[task].get('patience', self.patience)
            if self.tasks[task]['steps'] >= patience:
                # maybe stop entire training
                if is_target:
                    state_dict = torch.load(self.fid)
                    os.remove(self.fid)
                    raise EarlyStopException(task, self.tasks[task]['best'], state_dict)
                # update task weight
                else:
                    factor = self.tasks[task].get('factor', self.factor)
                    new_weight = self.tasks[task]['weight'] * factor
                    min_weight = self.tasks[task].get('min_weight', self.min_weight)
                    self.tasks[task]['weight'] = max(new_weight, min_weight)

    def get_weights(self):
        return {task: self.tasks[task]['weight'] for task in self.tasks}


class LRScheduler(object):
    def __init__(self, optimizer, threshold=0.0, **kwargs):
        self.lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, mode='max', threshold=threshold, **kwargs)

    def step(self, score):
        self.lr_scheduler.step(score)

    def __repr__(self):
        return '<LrScheduler lr="{:g}" steps="{}" patience="{}" threshold="{}"/>' \
            .format(self.lr_scheduler.optimizer.param_groups[0]['lr'],
                    self.lr_scheduler.num_bad_epochs,
                    self.lr_scheduler.patience,
                    self.lr_scheduler.threshold)


class Trainer(object):
    """
    Trainer

    Settings
    ========
    optim
    lr (lr_factor, lr_patience, min_lr)
    clip_norm
    weights
    report_freq
    checks_per_epoch
    """
    def __init__(self, settings, model, dataset, num_instances):
        self.target_task = get_target_task(settings)
        self.verbose = settings.verbose
        self.dataset = dataset
        self.model = model
        self.optimizer = getattr(optim, settings.optimizer)(
            model.parameters(), lr=settings.lr)
        self.clip_norm = settings.clip_norm

        self.report_freq = settings.report_freq
        self.num_batches = num_instances // dataset.batch_size
        if settings.checks_per_epoch == 1:
            self.check_freq = self.num_batches - 1  # check after last batch
        elif settings.checks_per_epoch > self.num_batches:
            self.check_freq = 1  # check after each batch
        elif settings.checks_per_epoch > 1:
            self.check_freq = self.num_batches // settings.checks_per_epoch  # check just
        else:
            self.check_freq = 0  # no checks

        self.task_scheduler = TaskScheduler(settings)
        self.lr_scheduler = LRScheduler(
            self.optimizer, factor=settings.lr_factor,
            patience=settings.lr_patience, min_lr=settings.min_lr)

        if settings.verbose:
            print()
            print("Evaluation check every {}/{} batches".format(
                self.check_freq, self.num_batches))
            print()
            print("::: Task schedules :::")
            print()
            print(self.task_scheduler)
            print()
            print("::: LR schedule :::")
            print()
            print(self.lr_scheduler)
            print()

    def weight_loss(self, loss):
        """
        Apply weights to losses and return a single loss number
        """
        weights = self.task_scheduler.get_weights()

        return sum(weights.get(k, 1) * loss[k] for k in loss)

    def evaluate(self, dataset):
        """
        Evaluate objective on held-out data
        """
        total_losses, total_batches = collections.defaultdict(float), 0

        # get all tasks
        tasks = list(self.model.tasks)

        for batch in tqdm.tqdm(dataset.batch_generator()):
            total_batches += 1
            for k, v in self.model.loss(batch, *tasks).items():
                total_losses[k] += v.item()

        for k, v in total_losses.items():
            total_losses[k] = v / total_batches

        return dict(total_losses)

    def run_check(self, devset):
        """
        Monitor dev loss and eventually early-stop training
        """
        print()
        print("Evaluating model on dev set...")
        print()

        self.model.eval()

        stored_scores = {}

        with torch.no_grad():
            dev_loss = self.evaluate(devset)
            print()
            print("::: Dev losses :::")
            print()
            print('\n'.join('{}: {:.4f}'.format(k, v) for k, v in dev_loss.items()))
            print()
            summary = self.model.evaluate(devset, self.dataset)
            for task_name, scorer in summary.items():
                stored_scores[task_name] = scorer.get_scores()
                scorer.print_summary(scores=stored_scores[task_name])

        self.model.train()
        dev_scores = {}
        for task, scored in stored_scores.items():
            dev_scores[task] = scored['all']['accuracy']
        # add lm scores
        if 'lm_fwd' in dev_loss or 'lm_bwd' in dev_loss:
            dev_scores['lm_fwd'] = dev_loss['lm_fwd']
            dev_scores['lm_bwd'] = dev_loss['lm_bwd']

        self.task_scheduler.step(dev_scores, self.model)
        self.lr_scheduler.step(dev_scores[self.target_task])

        if self.verbose:
            print(self.task_scheduler)
            print()
            print(self.lr_scheduler)
            print()

        return dev_scores

    def train_epoch(self, devset, epoch):
        rep_loss = collections.defaultdict(float)
        rep_batches = collections.defaultdict(int)
        rep_items, rep_start = 0, time.time()
        scores = None

        for b, batch in enumerate(self.dataset.batch_generator()):
            # get loss
            loss = self.model.loss(batch, get_batch_task(self.model.tasks.values()))

            if not loss:
                raise ValueError("Got empty loss, no tasks defined?")

            # optimize
            self.optimizer.zero_grad()
            self.weight_loss(loss).backward()
            if self.clip_norm > 0:
                clip_grad_norm_(self.model.parameters(), self.clip_norm)
            self.optimizer.step()

            # accumulate
            rep_items += type(self.dataset).get_nelement(batch)
            for k, v in loss.items():
                rep_batches[k] += 1
                rep_loss[k] += v.item()

            # report
            if b > 0 and b % self.report_freq == 0:
                rep = ""
                for t in sorted(rep_loss):
                    rep += '{}:{:.4f}  '.format(t, rep_loss[t] / rep_batches[t])
                logging.info("Batch [{}/{}] || {} || {:.0f} w/s".format(
                    b, self.num_batches, rep, rep_items / (time.time() - rep_start)))
                rep_loss = collections.defaultdict(float)
                rep_batches = collections.defaultdict(int)
                rep_items, rep_start = 0, time.time()

            if self.check_freq > 0 and b > 0 and b % self.check_freq == 0:
                if devset is not None:
                    rep_start = time.time()
                    scores = self.run_check(devset)
                    logging.info("Evaluation time: {:.0f} sec".format(
                        time.time() - rep_start))
                    rep_start = time.time()

        return scores

    def train_epochs(self, epochs, devset=None):
        """
        Train the model for a number of epochs
        """
        start = time.time()
        scores = None

        try:
            for epoch in range(1, epochs + 1):
                # train epoch
                epoch_start = time.time()
                logging.info("Starting epoch [{}]".format(epoch))
                self.train_epoch(devset, epoch)
                epoch_total = time.time() - epoch_start
                logging.info("Finished epoch [{}] in [{:.0f}] secs".format(
                    epoch, epoch_total))

        except EarlyStopException as e:
            logging.info("Early stopping training: "
                         "task [{}] with best score {:.4f}".format(e.task, e.loss))

            self.model.load_state_dict(e.best_state_dict)
            scores = {e.task: e.loss}

        logging.info("Finished training in [{:.0f}] secs".format(time.time() - start))

        # will be None if no dev test was provided or the model failed to converge
        return scores




import random
import json

import yaml
from json_minify import json_minify
import scipy.stats as stats

from pie import utils
from pie.settings import settings_from_file, check_settings, merge_task_defaults
from pie.settings import Settings


# available distributions
class truncnorm:
    def __init__(self, mu, std, lower=0, upper=1):
        a, b = (lower - mu) / std, (upper - mu) / std
        self.norm = stats.truncnorm(a, b, mu, std)

    def rvs(self):
        return float(self.norm.rvs())


class normint:
    def __init__(self, mu, std, lower, upper):
        self.norm = truncnorm(mu, std, lower, upper)

    def rvs(self):
        return int(round(self.norm.rvs())) // 2 * 2


class choice:
    def __init__(self, items):
        self.items = items

    def rvs(self):
        return random.choice(self.items)


def parse_opt(obj, opt_key):
    """
    Parses the opt file into a (possibly deep) dictionary where the leaves are 
    ready-to-use distributions
    """
    opt = {}

    for param, v in obj.items():
        if isinstance(v, list):
            opt[param] = [parse_opt(v_item, opt_key) for v_item in v]
        elif isinstance(v, dict):
            if opt_key in v:
                if v[opt_key] == 'norm':
                    opt[param] = stats.norm(**v['params'])
                elif v[opt_key] == 'truncnorm':
                    opt[param] = truncnorm(**v['params'])
                elif v[opt_key] == 'normint':
                    opt[param] = normint(**v['params'])
                elif v[opt_key] == 'choice':
                    opt[param] = choice(v['params'])
                else:
                    raise ValueError("Unknown distribution: ", v[opt_key])
            else:
                opt[param] = parse_opt(v, opt_key)
        else:
            opt[param] = v

    return opt


def read_opt(path, opt_key='opt'):
    """
    Reads and parses the opt file (as per parse_opt)
    """
    with open(path) as f:
        obj = json.loads(json_minify(f.read()))

    return parse_opt(obj, opt_key)


def sample_from_config(opt):
    """
    Applies the distributions specified in the opt.json file
    """
    output = {}

    for param, dist in opt.items():
        if isinstance(dist, dict):
            output[param] = sample_from_config(dist)
        elif isinstance(dist, list):
            output[param] = [sample_from_config(d) for d in dist]
        elif isinstance(dist, (str, float, int, bool)):
            output[param] = dist  # no sampling
        else:
            output[param] = dist.rvs()

    return output


def run_optimize(train_fn, settings, opt, n_iter, **kwargs):
    """
    Run random search over given `settings` resampling parameters as
    specified by `opt` for `n_iter` using `train_fn` function.

    - train_fn: a function that takes settings and any other possible kwargs
        and runs a training procedure
    - settings: a Settings object fully determining a training run
    - opt: a sampling file specifying parameters to resample each run,
        including a distribution to sample from. The contents are read from
        a json file with the following structure.
        { "lr": {
            "opt": "truncnorm",
            "params": {
                "mu": 0.0025, "std": 0.002, "lower": 0.0001, "upper": 1
                }
            }
        }
        "opt" specifies the distribution, and "params" the required parameters
        for that distribution:
            - "truncnorm": truncated normal
               - params: mu, std, lower, upper
            - "choice": uniform over given options
               - params: list of options
            - "normint": same as "truncnorm" but output is round up to an integer

        Other distributions can be implemented in the future.

    - n_iter: int, number of iterations to run
    """
    for i in range(n_iter):
        print()
        print("::: Starting optimization run {} :::".format(i + 1))
        print()
        sampled = sample_from_config(opt)
        merged = Settings(
            utils.recursive_merge(dict(settings), sampled, overwrite=True))
        print("::: Sampled settings :::")
        print(yaml.dump(dict(merged)))
        train_fn(check_settings(merge_task_defaults(merged)), **kwargs)


if __name__ == '__main__':
    from pie.settings import settings_from_file
    settings = settings_from_file("./transformer-lemma.json")
    opt = read_opt("opt-transformer.json")
    for _ in range(10):
        sampled = sample_from_config(opt)
        d = Settings(utils.recursive_merge(dict(settings), sampled, overwrite=True))
        for k in opt:
            print(k, d[k])
            print()




import os
import unittest
import copy

from pie.data.tabreader import TabReader


class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


class TestTabReader(unittest.TestCase):

    TASKS = [{"name": "lemma"}, {"name": "pos"}]
    FILENAME = '/tmp/test-data.tab'
    HEADER = {
        "header": False,
        "tasks_order": ["lemma", "pos", "ignore"],
        "sep": "\t",
        "breakline_ref": "input",
        "breakline_data": r"^(\.|''''|seg)$",
        "max_sent_len": 120,
        "tasks": TASKS
    }

    DEFAULT_INPUT = """SOIGNORS	seignor	NOMcom	NOMB.=p|GENRE=m|CAS=n
or	or4	ADVgen	DEGRE=-
escoutez	escouter	VERcjg	MODE=imp|PERS.=2|NOMB.=p
que	que4	CONsub	MORPH=empty
Dés	Dieu	NOMpro	NOMB.=s|GENRE=m|CAS=n
vos	vos	PROper	PERS.=2|NOMB.=p|GENRE=m|CAS=i
soit	estre1	VERcjg	MODE=sub|TEMPS=pst|PERS.=3|NOMB.=s
amis	ami	NOMcom	NOMB.=s|GENRE=m|CAS=n
{separator}
III	trois	DETcar	NOMB.=p|GENRE=m|CAS=r
vers	vers1	NOMcom	NOMB.=p|GENRE=m|CAS=r
de	de	PRE	MORPH=empty
bone	bon	ADJqua	NOMB.=s|GENRE=f|CAS=r|DEGRE=p
estoire	estoire1	NOMcom	NOMB.=s|GENRE=f|CAS=r
se	si	ADVgen	DEGRE=-
je	je	PROper	PERS.=1|NOMB.=s|GENRE=m|CAS=n
les	il	PROper	PERS.=3|NOMB.=p|GENRE=m|CAS=r
vos	vos	PROper	PERS.=2|NOMB.=p|GENRE=m|CAS=i
devis	deviser	VERcjg	MODE=ind|TEMPS=pst|PERS.=1|NOMB.=s
{separator}
Dou	de+le	PRE.DETdef	NOMB.=s|GENRE=m|CAS=r
premier	premier	ADJord	NOMB.=s|GENRE=m|CAS=r
roi	roi2	NOMcom	NOMB.=s|GENRE=m|CAS=r
de	de	PRE	MORPH=empty
France	France	NOMpro	NOMB.=s|GENRE=f|CAS=r
qui	qui	PROrel	NOMB.=s|GENRE=m|CAS=n
crestïens	crestiien	NOMcom	NOMB.=s|GENRE=m|CAS=n
devint	devenir	VERcjg	MODE=ind|TEMPS=psp|PERS.=3|NOMB.=s"""

    DIFF_INPUT = """SOIGNORS	seignor	NOMcom	NOMB.=p|GENRE=m|CAS=n
or	or4	ADVgen	DEGRE=-
escoutez	escouter	VERcjg	MODE=imp|PERS.=2|NOMB.=p
que	que4	CONsub	MORPH=empty
Dés	Dieu	NOMpro	NOMB.=s|GENRE=m|CAS=n
vos	vos	PROper	PERS.=2|NOMB.=p|GENRE=m|CAS=i
soit	estre1	VERcjg	MODE=sub|TEMPS=pst|PERS.=3|NOMB.=s
amis	ami	NOMcom	NOMB.=s|GENRE=m|CAS=n
{sep1}
III	trois	DETcar	NOMB.=p|GENRE=m|CAS=r
vers	vers1	NOMcom	NOMB.=p|GENRE=m|CAS=r
de	de	PRE	MORPH=empty
bone	bon	ADJqua	NOMB.=s|GENRE=f|CAS=r|DEGRE=p
estoire	estoire1	NOMcom	NOMB.=s|GENRE=f|CAS=r
se	si	ADVgen	DEGRE=-
je	je	PROper	PERS.=1|NOMB.=s|GENRE=m|CAS=n
les	il	PROper	PERS.=3|NOMB.=p|GENRE=m|CAS=r
vos	vos	PROper	PERS.=2|NOMB.=p|GENRE=m|CAS=i
devis	deviser	VERcjg	MODE=ind|TEMPS=pst|PERS.=1|NOMB.=s
{sep2}
Dou	de+le	PRE.DETdef	NOMB.=s|GENRE=m|CAS=r
premier	premier	ADJord	NOMB.=s|GENRE=m|CAS=r
roi	roi2	NOMcom	NOMB.=s|GENRE=m|CAS=r
de	de	PRE	MORPH=empty
France	France	NOMpro	NOMB.=s|GENRE=f|CAS=r
qui	qui	PROrel	NOMB.=s|GENRE=m|CAS=n
crestïens	crestiien	NOMcom	NOMB.=s|GENRE=m|CAS=n
devint	devenir	VERcjg	MODE=ind|TEMPS=psp|PERS.=3|NOMB.=s"""

    tokens = (
        ['SOIGNORS', 'or', 'escoutez', 'que', 'Dés', 'vos', 'soit', 'amis'],
        ['III', 'vers', 'de', 'bone', 'estoire', 'se', 'je', 'les', 'vos', 'devis'],
        ['Dou', 'premier', 'roi', 'de', 'France', 'qui', 'crestïens', 'devint']
    )

    def settings(self, **kwargs):
        x = {k+"": v for k, v in self.HEADER.items()}
        x.update(kwargs)
        return AttributeDict(x)

    def write(self, inp):
        with open(self.FILENAME, "w") as f:
            f.write(inp)

    def tearDown(self):
        if os.path.isfile(self.FILENAME):
            os.remove(self.FILENAME)

    def test_breakline_input(self):
        """ Regular expression should be taken into account as breaking line """
        self.write(self.DEFAULT_INPUT.format(separator="''''\tx\ty\tz"))

        tab_reader = TabReader(settings=self.settings(), fpath=self.FILENAME)

        seen = 0
        expected_tokens = copy.deepcopy(self.tokens)
        expected_tokens[0].append("''''")
        expected_tokens[1].append("''''")

        for (tokens, tasks), expected in zip(tab_reader.parselines(), expected_tokens):
            self.assertEqual(expected, tokens, "Sentence should be correctly cut")
            seen += 1

        self.assertEqual(seen, 3, "There should have been 3 sentences read")

    def test_breakline_input_empty_combination(self):
        """ Regular expression and empty lines should be taken into account as
        breaking line """

        self.write(self.DIFF_INPUT.format(sep1="''''\tx\ty\tz", sep2=""))

        tab_reader = TabReader(settings=self.settings(), fpath=self.FILENAME)

        seen = 0
        expected_tokens = copy.deepcopy(self.tokens)
        expected_tokens[0].append("''''")
        for (tokens, tasks), expected in zip(tab_reader.parselines(), expected_tokens):
            self.assertEqual(expected, tokens, "Sentence should be correctly cut")
            seen += 1

        self.assertEqual(seen, 3, "There should have been 3 sentences read")

    def test_breakline_empty_line(self):
        """ Empty lines should be taken into account as breaking line """
        self.write(self.DEFAULT_INPUT.format(separator=""))
        tab_reader = TabReader(settings=self.settings(), fpath=self.FILENAME)

        seen = 0
        for (tokens, tasks), expected in zip(tab_reader.parselines(), self.tokens):
            self.assertEqual(expected, tokens, "Sentence should be correctly cut")
            seen += 1

        self.assertEqual(seen, 3, "There should have been 3 sentences read")




import torch
import torch.nn as nn
import torch.nn.functional as F

from pie import torch_utils, initialization

from .embedding import build_embeddings
from .decoder import AttentionalDecoder, LinearDecoder, CRFDecoder
from .encoder import RNNEncoder
from .base_model import BaseModel


def get_context(outs, wemb, wlen, context_type):
    if context_type.lower() == 'sentence':
        return torch_utils.flatten_padded_batch(outs, wlen)
    elif context_type.lower() == 'word':
        return torch_utils.flatten_padded_batch(wemb, wlen)
    elif context_type.lower() == 'both':
        outs = torch_utils.flatten_padded_batch(outs, wlen)
        wemb = torch_utils.flatten_padded_batch(wemb, wlen)
        return torch.cat([outs, wemb], -1)
    else:
        return None


class SimpleModel(BaseModel):
    """
    Parameters
    ==========
    label_encoder : MultiLabelEncoder
    tasks : settings.tasks
    wemb_dim : int, embedding dimension for word-level embedding layer
    cemb_dim : int, embedding dimension for char-level embedding layer
    hidden_size : int, hidden_size for all hidden layers
    dropout : float
    merge_type : str, one of "concat", "mixer", method to merge word-level and
        char-level embeddings
    cemb_type : str, one of "RNN", "CNN", layer to use for char-level embeddings
    """
    def __init__(self, label_encoder, tasks, wemb_dim, cemb_dim, hidden_size, num_layers,
                 cell='LSTM', init_rnn='xavier_uniform',
                 # dropout
                 dropout=0.0, word_dropout=0.0,
                 # word embeddings
                 merge_type='concat', cemb_type='RNN', cemb_layers=1,
                 custom_cemb_cell=False,
                 # lm joint loss
                 include_lm=True, lm_shared_softmax=True,
                 # decoder
                 scorer='general', linear_layers=1,
                 # kwargs
                 **kwargs):
        # args
        self.wemb_dim = wemb_dim
        self.cemb_dim = cemb_dim
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        # kwargs
        self.cell = cell
        self.dropout = dropout
        self.word_dropout = word_dropout
        self.merge_type = merge_type
        self.cemb_type = cemb_type
        self.cemb_layers = cemb_layers
        self.custom_cemb_cell = custom_cemb_cell
        self.include_lm = include_lm
        self.lm_shared_softmax = lm_shared_softmax
        self.scorer = scorer
        self.linear_layers = linear_layers
        # only during training
        self.init_rnn = init_rnn
        super().__init__(label_encoder, tasks)

        # Embeddings
        (self.wemb, self.cemb, self.merger), in_dim = build_embeddings(
            label_encoder, wemb_dim,
            cemb_dim, cemb_type, custom_cemb_cell, cemb_layers, cell, init_rnn,
            merge_type, dropout)

        # Encoder
        self.encoder = None
        needs_encoder = False
        for task in self.tasks.values():
            if task['level'] == 'token':
                needs_encoder = True
                break
            elif task.get('context', '').lower() in ('sentence', 'both'):
                needs_encoder = True
                break
        if not needs_encoder:
            print("Model doesn't need sentence encoder, leaving uninitialized")
        else:
            self.encoder = RNNEncoder(
                in_dim, hidden_size, num_layers=num_layers, cell=cell, dropout=dropout,
                init_rnn=init_rnn)

        # Decoders
        decoders = {}
        for tname, task in self.tasks.items():

            if task['level'].lower() == 'char':
                if task['decoder'].lower() == 'linear':
                    decoder = LinearDecoder(
                        label_encoder.tasks[tname], self.cemb.embedding_dim)
                elif task['decoder'].lower() == 'crf':
                    decoder = CRFDecoder(
                        label_encoder.tasks[tname], self.cemb.embedding_dim)
                elif task['decoder'].lower() == 'attentional':
                    context_dim = 0
                    if task['context'].lower() == 'sentence':
                        context_dim = hidden_size * 2  # bidirectional encoder
                    elif task['context'].lower() == 'word':
                        context_dim = wemb_dim
                    elif task['context'].lower() == 'both':
                        context_dim = hidden_size * 2 + wemb_dim
                    decoder = AttentionalDecoder(
                        label_encoder.tasks[tname], cemb_dim, self.cemb.embedding_dim,
                        context_dim=context_dim, scorer=scorer, num_layers=cemb_layers,
                        cell=cell, dropout=dropout, init_rnn=init_rnn)
                else:
                    raise ValueError(
                        "Unknown decoder type {} for char-level task: {}".format(
                            task['decoder'], tname))

            elif task['level'].lower() == 'token':
                # linear
                if task['decoder'].lower() == 'linear':
                    decoder = LinearDecoder(
                        label_encoder.tasks[tname], hidden_size * 2, dropout=dropout,
                        highway_layers=linear_layers - 1)
                # crf
                elif task['decoder'].lower() == 'crf':
                    decoder = CRFDecoder(
                        label_encoder.tasks[tname], hidden_size * 2,
                        highway_layers=linear_layers - 1)

            else:
                raise ValueError(
                    "Unknown decoder type {} for token-level task: {}".format(
                        task['decoder'], tname))

            self.add_module('{}_decoder'.format(tname), decoder)
            decoders[tname] = decoder

        self.decoders = decoders

        # - LM
        if self.include_lm:
            self.lm_fwd_decoder = LinearDecoder(label_encoder.word, hidden_size)
            if lm_shared_softmax:
                self.lm_bwd_decoder = self.lm_fwd_decoder
            else:
                self.lm_bwd_decoder = LinearDecoder(label_encoder.word, hidden_size)

    def get_args_and_kwargs(self):
        return {'args': (self.wemb_dim, self.cemb_dim,
                         self.hidden_size, self.num_layers),
                'kwargs': {'dropout': self.dropout,
                           'word_dropout': self.word_dropout,
                           'cell': self.cell,
                           'merge_type': self.merge_type,
                           'linear_layers': self.linear_layers,
                           'cemb_type': self.cemb_type,
                           'cemb_layers': self.cemb_layers,
                           'include_lm': self.include_lm,
                           'scorer': self.scorer,
                           'custom_cemb_cell': self.custom_cemb_cell}}

    def embedding(self, word, wlen, char, clen):
        wemb, cemb, cemb_outs = None, None, None
        if self.wemb is not None:
            # set words to unknown with prob `p` depending on word frequency
            word = torch_utils.word_dropout(
                word, self.word_dropout, self.training, self.label_encoder.word)
            wemb = self.wemb(word)
        if self.cemb is not None:
            # cemb_outs: (seq_len x batch x emb_dim)
            cemb, cemb_outs = self.cemb(char, clen, wlen)

        if wemb is None:
            emb = cemb
        elif cemb is None:
            emb = wemb
        else:
            emb = self.merger(wemb, cemb)

        return emb, (wemb, cemb, cemb_outs)

    def init_from_encoder(self, encoder):
        # wemb
        total = 0
        for w, idx in encoder.label_encoder.word.table.items():
            if w in self.label_encoder.word.table:
                self.wemb.weight.data[self.label_encoder.word.table[w]].copy_(
                    encoder.wemb.weight.data[idx])
                total += 1
        print("Initialized {}/{} word embs".format(total, len(self.wemb.weight)))
        # cemb
        total = 0
        for w, idx in encoder.label_encoder.char.table.items():
            if w in self.label_encoder.char.table:
                self.cemb.emb.weight.data[self.label_encoder.char.table[w]].copy_(
                    encoder.cemb.emb.weight.data[idx])
                total += 1
        print("Initialized {}/{} char embs".format(total, len(self.cemb.emb.weight)))
        # cemb rnn
        self.cemb.rnn.load_state_dict(encoder.cemb.rnn.state_dict())
        # sentence rnn
        self.encoder.load_state_dict(encoder.encoder.state_dict())

        if self.include_lm:
            pass

    def loss(self, batch_data, *target_tasks):
        ((word, wlen), (char, clen)), tasks = batch_data
        output = {}

        # Embedding
        emb, (wemb, cemb, cemb_outs) = self.embedding(word, wlen, char, clen)

        # Encoder
        emb = F.dropout(emb, p=self.dropout, training=self.training)
        enc_outs = None
        if self.encoder is not None:
            # TODO: check if we need encoder for this particular batch
            enc_outs = self.encoder(emb, wlen)

        # Decoder
        for task in target_tasks:
            (target, length), at_layer = tasks[task], self.tasks[task]['layer']
            # prepare input layer
            outs = None
            if enc_outs is not None:
                outs = F.dropout(enc_outs[at_layer], p=0, training=self.training)

            decoder = self.decoders[task]

            if self.tasks[task]['level'].lower() == 'char':
                if isinstance(decoder, LinearDecoder):
                    logits = decoder(cemb_outs)
                    output[task] = decoder.loss(logits, target)
                elif isinstance(decoder, CRFDecoder):
                    logits = decoder(cemb_outs)
                    output[task] = decoder.loss(logits, target, length)
                elif isinstance(decoder, AttentionalDecoder):
                    cemb_outs = F.dropout(
                        cemb_outs, p=self.dropout, training=self.training)
                    context = get_context(outs, wemb, wlen, self.tasks[task]['context'])
                    logits = decoder(target, length, cemb_outs, clen, context=context)
                    output[task] = decoder.loss(logits, target)
            else:
                if isinstance(decoder, LinearDecoder):
                    logits = decoder(outs)
                    output[task] = decoder.loss(logits, target)
                elif isinstance(decoder, CRFDecoder):
                    logits = decoder(outs)
                    output[task] = decoder.loss(logits, target, length)

        # (LM)
        if self.include_lm:
            if len(emb) > 1:  # can't compute loss for 1-length batches
                # always at first layer
                fwd, bwd = F.dropout(
                    enc_outs[0], p=0, training=self.training
                ).chunk(2, dim=2)
                # forward logits
                logits = self.lm_fwd_decoder(torch_utils.pad(fwd[:-1], pos='pre'))
                output['lm_fwd'] = self.lm_fwd_decoder.loss(logits, word)
                # backward logits
                logits = self.lm_bwd_decoder(torch_utils.pad(bwd[1:], pos='post'))
                output['lm_bwd'] = self.lm_bwd_decoder.loss(logits, word)

        return output

    def predict(self, inp, *tasks, return_probs=False,
                use_beam=False, beam_width=10, **kwargs):
        """
        inp : (word, wlen), (char, clen), text input
        tasks : list of str, target tasks
        """
        tasks = set(self.tasks if not len(tasks) else tasks)
        preds, probs = {}, {}
        (word, wlen), (char, clen) = inp

        # Embedding
        emb, (wemb, cemb, cemb_outs) = self.embedding(word, wlen, char, clen)

        # Encoder
        enc_outs = None
        if self.encoder is not None:
            # TODO: check if we need encoder for this particular batch
            enc_outs = self.encoder(emb, wlen)

        # Decoders
        for task in tasks:

            decoder, at_layer = self.decoders[task], self.tasks[task]['layer']
            outs = None
            if enc_outs is not None:
                outs = enc_outs[at_layer]

            if self.label_encoder.tasks[task].level.lower() == 'char':
                if isinstance(decoder, LinearDecoder):
                    hyps, prob = decoder.predict(cemb_outs, clen)
                elif isinstance(decoder, CRFDecoder):
                    hyps, prob = decoder.predict(cemb_outs, clen)
                else:
                    context = get_context(outs, wemb, wlen, self.tasks[task]['context'])
                    if use_beam:
                        hyps, prob = decoder.predict_beam(
                            cemb_outs, clen, width=beam_width, context=context)
                    else:
                        hyps, prob = decoder.predict_max(
                            cemb_outs, clen, context=context)
                    if self.label_encoder.tasks[task].preprocessor_fn is None:
                        hyps = [''.join(hyp) for hyp in hyps]
            else:
                if isinstance(decoder, LinearDecoder):
                    hyps, prob = decoder.predict(outs, wlen)
                elif isinstance(decoder, CRFDecoder):
                    hyps, prob = decoder.predict(outs, wlen)

            preds[task] = hyps
            probs[task] = prob

        if return_probs:
            return preds, probs

        return preds


if __name__ == '__main__':
    from pie.settings import settings_from_file
    from pie.data import Dataset, Reader, MultiLabelEncoder

    settings = settings_from_file('./config.json')
    reader = Reader(settings, settings.input_path)
    label_encoder = MultiLabelEncoder.from_settings(settings)
    label_encoder.fit_reader(reader)
    data = Dataset(settings, reader, label_encoder)
    model = SimpleModel(data.label_encoder, settings.tasks,
                        settings.wemb_dim, settings.cemb_dim,
                        settings.hidden_size, settings.num_layers)
    model.to(settings.device)

    for batch in data.batch_generator():
        model.loss(batch)
        break
    ((word, wlen), (char, clen)), tasks = next(data.batch_generator())
    wemb, (cemb, cemb_outs) = model.wemb(word), model.cemb(char, clen, wlen)
    emb = model.merger(wemb, cemb)
    enc_outs = model.encoder(emb, wlen)
    model.pos_decoder.predict(enc_outs, wlen)
    lemma_hyps, _ = model.decoders['lemma'].predict_max(
        cemb_outs, clen, context=torch_utils.flatten_padded_batch(enc_outs, wlen))
    print(lemma_hyps)




from .utils import GitInfo

try:
    __commit__ = GitInfo(__file__).get_commit()
except Exception:
    from .commit_build import COMMIT
    if COMMIT:
        __commit__ = COMMIT
    else:
        import logging
        logging.warning(
            """
It seems like you downloaded `pie` instead of git-cloning it or installing it with pip.
We won't be able to check compatibility between pretrained models and `pie` version.
""")
        __commit__ = None

from . import utils
from . import trainer
from . import settings
from . import tagger
from . import initialization
from . import optimize
from .data import *
from .models import *
from .pretrain_encoder import Encoder




import torch
import torch.nn as nn


def init_embeddings(embeddings):
    embeddings.reset_parameters()
    nn.init.constant_(embeddings.weight, 0.01)


def init_linear(linear):
    linear.reset_parameters()
    nn.init.xavier_uniform_(linear.weight)


def init_rnn(rnn, forget_bias=1.0, scheme='default'):
    print("Initializing {} with scheme: {}".format(type(rnn).__name__, scheme))
    for pname, p in rnn.named_parameters():
        if 'bias' in pname:
            nn.init.constant_(p, 0.)
            # forget_bias
            if 'LSTM' in type(rnn).__name__:
                nn.init.constant_(p[rnn.hidden_size:rnn.hidden_size*2], forget_bias)

        elif pname.startswith('weight_ih'):
            if scheme == 'xavier_uniform':
                nn.init.xavier_uniform_(p)
            elif scheme == 'orthogonal':
                nn.init.orthogonal_(p)

        elif pname.startswith('weight_hh'):
            if scheme == 'xavier_uniform':
                nn.init.xavier_uniform_(p)
            elif scheme == 'orthogonal':
                gates = 1
                if 'LSTM' in type(rnn).__name__:
                    gates = 4
                elif 'GRU' in type(rnn).__name__:
                    gates = 3
                for i in range(gates):
                    nn.init.eye_(p[i * rnn.hidden_size: (i+1) * rnn.hidden_size])


def init_conv(conv):
    conv.reset_parameters()
    nn.init.xavier_uniform_(conv.weight)
    nn.init.constant_(conv.bias, 0.)
    pass


def init_pretrained_embeddings(path, encoder, embedding):
    inits = 0
    with open(path) as f:
        # maybe validate
        header = next(f)
        if len(header.split()) > 2:
            print("Skipping header validation for embedding file: {}".format(path))
            f = (line for it in [[header], f] for line in it)
        else:
            nemb, dim = next(f).split()
            if int(dim) != embedding.weight.data.size(1):
                raise ValueError("Unexpected embeddings size: {}".format(dim))

        for line in f:
            word, *vec = line.split()
            if word in encoder.table:
                embedding.weight.data[encoder.table[word], :].copy_(
                    torch.tensor([float(v) for v in vec]))
                inits += 1

    if embedding.padding_idx is not None:
        embedding.weight.data[embedding.padding_idx].zero_()

    print("Initialized {}/{} embeddings".format(inits, embedding.num_embeddings))




import torch
import torch.nn as nn
import torch.nn.functional as F

from pie import initialization


class Highway(nn.Module):
    """
    Highway network
    """
    def __init__(self, in_features, num_layers, act='relu'):
        self.in_features = in_features

        self.act = act
        super().__init__()

        self.layers = nn.ModuleList(
            [nn.Linear(in_features, in_features*2) for _ in range(num_layers)])

        self.init()

    def init(self):
        for layer in self.layers:
            initialization.init_linear(layer)
            # bias gate to let information go untouched
            nn.init.constant_(layer.bias[self.in_features:], 1.)

    def forward(self, inp):
        current = inp
        for layer in self.layers:
            inp, gate = layer(current).chunk(2, dim=-1)
            inp, gate = getattr(F, self.act)(inp), torch.sigmoid(gate)
            current = gate * current + (1 - gate) * inp

        return current






from .edit_trees import make_edit_tree


def transform(lem, tok):
    return make_edit_tree(tok, lem).to_tuple()


def inverse_transform(pred, tok):
    plen, pstring, slen, sstring = pred
    return pstring + tok[plen:-slen or None] + sstring




import torch
import torch.nn as nn

from pie import torch_utils
from pie import initialization


class RNNEncoder(nn.Module):
    def __init__(self, in_size, hidden_size, num_layers=1, cell='GRU', dropout=0.0,
                 init_rnn='default'):

        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.cell = cell
        self.dropout = dropout
        self.init_rnn = init_rnn
        super().__init__()

        rnn = []
        for layer in range(num_layers):
            rnn_inp = in_size if layer == 0 else hidden_size * 2
            rnn.append(getattr(nn, cell)(rnn_inp, hidden_size, bidirectional=True))
        self.rnn = nn.ModuleList(rnn)

        self.init()

    def init(self):
        for rnn in self.rnn:
            initialization.init_rnn(rnn, scheme=self.init_rnn)

    def forward(self, inp, lengths, hidden=None, only_last=False, return_hidden=False):
        if hidden is None:
            hidden = [
                torch_utils.init_hidden_for(
                    inp, 2, 1, self.hidden_size, self.cell, add_init_jitter=False)
                for _ in range(len(self.rnn))]

        _, sort = torch.sort(lengths, descending=True)
        _, unsort = sort.sort()
        inp = nn.utils.rnn.pack_padded_sequence(inp[:, sort], lengths[sort].cpu())
        outs, hiddens = [], []

        for layer, rnn in enumerate(self.rnn):
            # apply dropout only in between layers (not on the output)
            if layer > 0:
                inp, lengths = nn.utils.rnn.pad_packed_sequence(inp)
                inp = torch_utils.sequential_dropout(inp, self.dropout, self.training)
                inp = nn.utils.rnn.pack_padded_sequence(inp, lengths.cpu())
            # run layer
            louts, lhidden = rnn(inp, hidden[layer])
            # unpack
            louts_, _ = nn.utils.rnn.pad_packed_sequence(louts)
            outs.append(louts_[:, unsort])
            if isinstance(lhidden, tuple):
                lhidden = lhidden[0][:, unsort, :], lhidden[1][:, unsort, :]
            else:
                lhidden = lhidden[:, unsort, :]
            hiddens.append(lhidden)
            # recur
            inp = louts

        if only_last:
            outs, hiddens = outs[-1], hiddens[-1]

        if return_hidden:
            return outs, hiddens
        else:
            return outs




import re
import collections

from lxml import etree

from .base_reader import BaseReader, LineParseException


TEI = 'http://www.tei-c.org/ns/1.0'


def get_tei_tag(item):
    "Scrap namespace from tag"
    return re.sub("{{{}}}".format(TEI), "", item.tag)


def parse_pos(pos, simplify=False):
    "eventually simplify the pos tag removing subclass specs"
    if simplify:
        return re.sub(r"\([^\)]+\)", "", pos)
    return pos


class TEIReader(BaseReader):
    """
    TEI reader for BaB files
    """
    def __init__(self, settings, fpath):
        super(TEIReader, self).__init__(settings, fpath)

        self.max_sent_len = settings.max_sent_len

    def parselines(self):
        tree = etree.parse(self.fpath)
        inp, tasks_data = [], collections.defaultdict(list)

        for num, item in enumerate(tree.find('//tei:s', namespaces={'tei': TEI})):
            # avoid too long sentences
            if len(inp) >= self.max_sent_len:
                yield inp, dict(tasks_data)
                inp, tasks_data = [], collections.defaultdict(list)

            # check item tag
            tag = get_tei_tag(item)

            # ignore (physical line-breaks, deletions, editorials)
            if tag in ('lb', 'del', 'seg'):
                continue

            # punctuation
            elif tag == 'pc':
                inp.append(item.text.strip())
                tasks_data['pos'].append('PUNC')
                tasks_data['lemma'].append(item.text.strip())
                # # TODO: <pc> should segment nice sentences but it doesn't seem to
                # yield inp, dict(tasks_data)
                # inp, tasks_data = [], collections.defaultdict(list)

            # words
            elif tag == 'w':
                if item.text is None:
                    # <w lemma="" pos="" misAlignment="true:1/0" changed="yes"/>
                    continue
                inp.append(item.text.strip())
                pos = parse_pos(item.attrib['pos'], simplify=True)
                tasks_data['pos'].append(pos)
                lemma = item.attrib['lemma']
                tasks_data['lemma'].append(lemma)

            # unknown
            else:
                yield LineParseException(
                    "Encountered unexpected tag: {} at position #{}, file {}".format(
                        tag, num, self.fpath))
                inp, tasks_data = [], collections.defaultdict(list)

        # yield rest
        if len(inp) > 0:
            yield inp, tasks_data

    def get_tasks(self):
        return ('lemma', 'pos')




# Can be run with python -m pie.scripts.tag_pipe
import sys
from pie import utils
from pie.tagger import Tagger, simple_tokenizer


def run(model_spec, device, batch_size, lower, beam_width, use_beam, tokenize):
    with utils.shutup():
        tagger = Tagger(device=device, batch_size=batch_size, lower=lower)

        for model, tasks in model_spec:
            tagger.add_model(model, *tasks)
            tasks = tasks or tagger.models[-1][0].label_encoder.tasks

    header = False
    for line in sys.stdin:
        if not line:
            continue

        if tokenize:
            line = simple_tokenizer(line, lower)
        else:
            line = line.split()

        preds, tasks = tagger.tag(
            [line], [len(line)], use_beam=use_beam, beam_width=beam_width)

        if not header:
            print('\t'.join(['token'] + tasks))
            header = True

        preds = preds[0]  # unpack
        tokens, tags = zip(*preds)
        for token, tags in zip(tokens, tags):
            print('\t'.join([token] + list(tags)))


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('model_spec', type=utils.model_spec)
    parser.add_argument('--device', default='cpu')
    parser.add_argument('--use_beam', action='store_true')
    parser.add_argument('--batch_size', type=int, default=50)
    parser.add_argument('--beam_width', default=10, type=int)
    parser.add_argument('--lower', action='store_true')
    parser.add_argument('--tokenize', action='store_true')
    args = parser.parse_args()
    run(model_spec=args.model_spec, device=args.device,
        batch_size=args.batch_size,
        lower=args.lower, beam_width=args.beam_width,
        use_beam=args.use_beam, tokenize=args.tokenize)




import numpy as np


def get_pretrained_embeddings(reader, label_encoder, **kwargs):
    from pie import utils
    with utils.shutup():        # avoid pattern warning
        from gensim.models import Word2Vec

    word2vec = Word2Vec(reader.get_token_iterator(), **kwargs)
    weight = np.zeros((len(label_encoder.word), word2vec.wv.vector_size))

    found = 0
    for w, idx in label_encoder.word.table.items():
        try:
            weight[idx] = word2vec.wv[w]
            found += 1
        except KeyError:  # reserved symbols are not in training sentences
            pass

    print("A total of {}/{} word embeddings were pretrained"
          .format(found, len(label_encoder.word)))

    return weight


if __name__ == '__main__':
    from pie.data import Reader, MultiLabelEncoder
    from pie.settings import settings_from_file

    settings = settings_from_file("config.json")
    reader = Reader(settings, settings.input_path)
    le = MultiLabelEncoder.from_settings(settings)
    le.fit_reader(reader)
    get_pretrained_embeddings(reader, le, min_count=1)




import os
import random

from pie.utils import get_filenames

from .tabreader import TabReader
from .conll_reader import CONLLReader, CONLLUReader
from .tei_reader import TEIReader


class Reader(object):
    """
    Dispatcher class

    Parameters
    ==========
    input_path : str (optional), either a path to a directory, a path to a file
        or a unix style pathname pattern expansion for glob

    Settings
    ========
    shuffle : bool, whether to shuffle files after each iteration
    max_sents : int, maximum number of sentences to read (note that depending
        on shuffle the result might be non-determinitic)
    """
    def __init__(self, settings, *input_paths):
        filenames = []
        for input_path in input_paths:
            if input_path is not None:
                filenames.extend(get_filenames(input_path))

        if len(filenames) == 0:
            raise RuntimeError("Couldn't find files [{}]".format(''.join(input_paths)))

        self.readers = [self.get_reader(fpath)(settings, fpath) for fpath in filenames]

        # settings
        self.shuffle = settings.shuffle
        self.max_sents = settings.max_sents
        # cache
        self.nsents = None

    def get_reader(self, fpath):
        """
        Decide on reader type based on filename
        """
        # imports

        if fpath.endswith('tab') or fpath.endswith('tsv') or fpath.endswith('csv'):
            return TabReader

        elif fpath.endswith('conll'):
            return CONLLReader

        elif fpath.endswith('conllu'):
            return CONLLUReader

        elif fpath.endswith('xml'):
            return TEIReader

        else:
            raise ValueError("Unknown file format: {}".format(fpath))

    def reset(self):
        """
        Called after a full run over `readsents`
        """
        if self.shuffle:
            random.shuffle(self.readers)

    def check_tasks(self, expected=None):
        """
        Check tasks over files
        """
        tasks = set()

        for reader in self.readers:
            tasks.update(reader.check_tasks(expected=expected))

        return tuple(tasks)

    def readsents(self, silent=True, only_tokens=False):
        """
        Read sents over files
        """
        self.reset()
        total = 0
        for reader in self.readers:
            for data in reader.readsents(silent=silent, only_tokens=only_tokens):
                # check # lines processed
                if total >= self.max_sents:
                    break
                total += 1

                yield data
        self.nsents = total

    def get_nsents(self):
        """
        Number of sents in Reader
        """
        if self.nsents is not None:
            return self.nsents
        nsents = 0
        for _ in self.readsents():
            nsents += 1
        self.nsents = nsents
        return nsents

    def get_token_iterator(self):
        """
        Get an iterator over sentences of plain tokens
        """
        return TokenIterator(self)


class TokenIterator():
    def __init__(self, reader):
        self.reader = reader

    def __iter__(self):
        yield from self.reader.readsents(only_tokens=True)




import re
import io
import os
import shutil
import uuid
import gzip
import tarfile
import logging
import sys
import glob
import itertools
import functools
import unicodedata
from contextlib import contextmanager
from subprocess import check_output, CalledProcessError
import tempfile


DEFAULT_TEMP_DIR = tempfile.gettempdir()


def lower_str(s):
    return s.lower()


def apply_utfnorm(s, form='NFKD'):
    return unicodedata.normalize(form, s)


def drop_diacritics(s):
    return ''.join(c for c in unicodedata.normalize('NFD', s)
                   if unicodedata.category(c) != 'Mn')


def identity(*args):
    if len(args) > 1:
        return args
    return args[0]


def compose(*fns):
    def compose_fn(f, g):
        def fn(*args, **kwargs):
            return f(g(*args, **kwargs))
        return fn
    return functools.reduce(compose_fn, fns, identity)


def window(it):
    """
    >>> list(window(range(5)))
    [(None, 0, 1), (0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, None)]
    """
    it = itertools.chain([None], it, [None])  # pad for completeness
    result = tuple(itertools.islice(it, 3))

    if len(result) == 3:
        yield result

    for elem in it:
        result = result[1:] + (elem,)
        yield result


def chunks(it, size):
    """
    Chunk a generator into a given size (last chunk might be smaller)
    """
    buf = []
    for s in it:
        buf.append(s)
        if len(buf) == size:
            yield buf
            buf = []
    if len(buf) > 0:
        yield buf


def flatten(it):
    """
    >>> list(flatten([['abc', 'cde'], ['yte']]))
    ['a', 'b', 'c', 'c', 'd', 'e', 'y', 't', 'e']
    """
    if isinstance(it, str):
        for i in it:
            yield i
    else:
        for subit in it:
            yield from flatten(subit)


def recursive_merge(s1, s2, overwrite=False):
    """
    Recursively merge two dictionaries

    >>> recursive_merge({"a": {"b": 1}}, {"a": {"c": 2}})
    {'a': {'b': 1, 'c': 2}}
    """
    for k, v in s2.items():
        if k in s1 and isinstance(v, dict):
            if not isinstance(s1[k], dict):
                raise ValueError("Expected dictionary at key [{}]".format(k))
            s1[k] = recursive_merge(s1[k], v, overwrite=overwrite)
        else:
            if overwrite or k not in s1:
                s1[k] = v

    return s1


def ensure_ext(path, ext, infix=None):
    """
    Compute target path with eventual infix and extension

    >>> ensure_ext("model.pt", "pt", infix="0.87")
    'model-0.87.pt'
    >>> ensure_ext("model.test", "pt", infix="0.87")
    'model-0.87.test.pt'
    >>> ensure_ext("model.test", "test", infix="pie")
    'model-pie.test'
    """
    path, oldext = os.path.splitext(path)

    # normalize extension
    if ext.startswith("."):
        ext = ext[1:]
    if oldext.startswith("."):
        oldext = oldext[1:]

    # infix
    if infix is not None:
        path = "-".join([path, infix])

    # add old extension if not the same as the new one
    if oldext and oldext != ext:
        path = '.'.join([path, oldext])

    return '.'.join([path, ext])


def get_filenames(input_path):
    """
    Get filenames from path expression
    """
    if os.path.isdir(input_path):
        filenames = [os.path.join(input_path, f) for f in os.listdir(input_path)
                     if not f.startswith('.')]
    elif os.path.isfile(input_path):
        filenames = [input_path]
    else:                       # unix string
        filenames = glob.glob(input_path)

    return filenames


@contextmanager
def shutup():
    with open(os.devnull, "w") as void:
        old_out = sys.stdout
        old_err = sys.stderr
        sys.stdout = void
        sys.stderr = void
        try:
            yield
        finally:
            sys.stdout = old_out
            sys.stderr = old_err


@contextmanager
def tmpfile(parent=DEFAULT_TEMP_DIR):
    fid = str(uuid.uuid1())
    tmppath = os.path.join(parent, fid)
    yield tmppath
    if os.path.isdir(tmppath):
        shutil.rmtree(tmppath)
    else:
        os.remove(tmppath)


def add_weights_to_tar(state_dict, path, tar):
    import torch
    f = io.BytesIO()
    torch.save(state_dict, f)
    tinf = tarfile.TarInfo(name=path)
    f.seek(0)
    tinf.size = len(f.getbuffer())
    tar.addfile(tinf, f)     # read tinf.size bytes from f into tinf


def add_gzip_to_tar(string, subpath, tar):
    with tmpfile() as tmppath:
        with gzip.GzipFile(tmppath, 'w') as f:
            f.write(string.encode())
        tar.add(tmppath, arcname=subpath)


def get_gzip_from_tar(tar, fpath):
    return gzip.open(tar.extractfile(fpath)).read().decode().strip()


class GitInfo():
    """
    Utility class to retrieve git-based info from a repository
    """
    def __init__(self, fname):
        if os.path.isfile(fname):
            self.dirname = os.path.dirname(fname)
        elif os.path.isdir(fname):
            self.dirname = fname
        else:
            # not a file
            self.dirname = None

        if not os.path.isfile(fname) and not os.path.isdir(fname):
            logging.warn("[GitInfo]: Input file doesn't exit")

        else:
            try:
                with shutup():
                    check_output(['git', '--version'], cwd=self.dirname)
            except FileNotFoundError:
                self.dirname = None
                logging.warn("[GitInfo]: Git doesn't seem to be installed")
            except CalledProcessError as e:
                self.dirname = None
                code, _ = e.args
                if code == 128:
                    logging.warn("[GitInfo]: Script not git-tracked")
                else:
                    logging.warn("[GitInfo]: Unrecognized git error")

    def run(self, cmd):
        if self.dirname is None:
            return

        return check_output(cmd, cwd=self.dirname).strip().decode('utf-8')

    def get_commit(self):
        """
        Returns current commit on file or None if an error is thrown by git
        (OSError) or if file is not under git VCS (CalledProcessError)
        """
        return self.run(["git", "describe", "--always"])

    def get_branch(self):
        """
        Returns current active branch on file or None if an error is thrown
        by git (OSError) or if file is not under git VCS (CalledProcessError)
        """
        return self.run(["git", "rev-parse", "--abbrev-ref", "HEAD"])

    def get_tag(self):
        """
        Returns current active tag
        """
        return self.run(["git", "describe", "--tags", "--abbrev=0"])


def model_spec(inp):
    """
    >>> example = 'model-pos-2018:03:05.tar'
    >>> model_spec(example)
    [('model-pos-2018:03:05.tar', [])]

    >>> example = '<model-pos-2018:03:05.tar,pos><model-pos-2018:03:05.tar,lemma>'
    >>> model_spec(example)
    [('model-pos-2018:03:05.tar', ['pos']), ('model-pos-2018:03:05.tar', ['lemma'])]
    """
    if not inp.startswith('<'):
        return [(inp, [])]

    output = []
    for string in re.findall(r'<([^>]+)>', inp):
        model_path, *tasks = string.split(',')
        output.append((model_path, tasks))

    return output




import collections

from .base_reader import BaseReader, LineParseException


PROIELMORPH = {
    "CASE": "case",
    "DEGR": "degree",
    "GEND": "gender",
    "INFL": None,               # only has one possible value (n)
    "MOOD": "mood",
    "NUMB": "number",
    "PERS": "person",
    "STRE": None,               # no idea what this is
    "TENS": "tense",
    "VOIC": "voice"
}


def get_lines(fpath, _parse_morph):
    with open(fpath) as f:
        for line_num, line in enumerate(f):
            line = line.strip()
            if not line:
                continue
            elif line.startswith('#'):
                continue
            else:
                num, form, lemma, pos, ppos, morph, _, dep, *_ = line.split('\t')
                tasks = {'lemma': lemma, 'pos': pos, 'ppos': ppos, 'dep': dep}
                for task, val in _parse_morph(morph).items():
                    tasks[task] = val
                yield form, tasks


def get_sents(fpath, parse_morph_):
    lines = 0
    with open(fpath) as f:
        sent, prev, tasks = [], 0, collections.defaultdict(list)
        for line in f:
            line = line.strip()

            if not line:
                # new line
                yield sent, dict(tasks)
                sent, prev, tasks = [], 0, collections.defaultdict(list)
                lines += 1

            elif line.startswith('#'):
                # skip metadata
                continue

            else:
                num, form, lemma, pos, ppos, morph, _, dep, *_ = line.split('\t')
                try:
                    num = int(num)
                    assert num == prev + 1, (num, prev)
                    prev = num

                    sent.append(form)
                    tasks['lemma'].append(lemma)
                    tasks['pos'].append(pos)
                    tasks['ppos'].append(ppos)
                    tasks['dep'].append(dep)
                    for key, val in parse_morph_(morph).items():
                        tasks[key].append(val)

                except ValueError:
                    # 20-22	"průběžná_inventarizace"
                    continue


class CONLLReader(BaseReader):
    """
    CONLLReader format as found on the PROIEL files

    6	David	David	N	Ne	INFLn	5	atr	_	_

    which corresponds to:

    LINE_NUM	FORM	LEMMA	POS	PPOS	MORPH	HEAD	DEP	_	_
    """
    def __init__(self, settings, fpath):
        super(CONLLReader, self).__init__(settings, fpath)
        self.max_sent_len = settings.max_sent_len

    def _parse_morph(self, morph):
        if morph == '_':
            return morph
        output = {}
        for val in morph.split("|"):
            task, val = val[:-1], val[-1]
            if PROIELMORPH.get(task):
                output[PROIELMORPH[task]] = val
        return output

    def parse_morph_(self, morph):
        morph = self._parse_morph(morph)
        output = {}
        for task in self.tasks:
            if task not in ('lemma', 'pos', 'ppos', 'dep'):
                output[task] = morph.get(task, '_')
        return output

    def parselines(self):
        """
        Generator over sentences in a single file
        """
        inp = []

        for inp, tasks in get_sents(self.fpath, self.parse_morph_):
            while len(inp) > self.max_sent_len:
                inp_ = inp[:self.max_sent_len]
                tasks_ = {}
                for task in tasks:
                    tasks_[task] = tasks[task][:self.max_sent_len]
                yield inp_, tasks_
                inp = inp[self.max_sent_len:]
                for task in tasks:
                    tasks[task] = tasks[task][self.max_sent_len:]
            yield inp, tasks

    def get_tasks(self):
        """
        All conll tasks (as in proiel files) in expected order
        """
        output = set()
        for _, tasks in get_lines(self.fpath, self._parse_morph):
            for task in tasks:
                output.add(task)

        return tuple(output)


class CONLLUReader(CONLLReader):
    def _parse_morph(self, morph):
        if morph == '_':
            return {}
        output = {}
        for val in morph.split("|"):
            key, val = val.split("=")
            output[key] = val
        return output




import re
import string
import os
import itertools

import tqdm

from pie.models import BaseModel
from pie.data import pack_batch
from pie import utils


def regexsplitter(regex):
    def func(line):
        for line in re.split(regex, line):
            line = line.strip()
            if line:
                yield line

    return func


SECTION = r'([A-Z\. ]+\.)'
FULLSTOP = r'([^\.]+\.)'
WORD = r'([{}])'.format(string.punctuation)


def simple_tokenizer(text):
    section = regexsplitter(SECTION)
    fullstop = regexsplitter(FULLSTOP)
    word = regexsplitter(WORD)

    for line in section(text):
        for sent in fullstop(line):
            yield [w for raw in sent.split() for w in word(raw)]


def get_sentences_vrt(fpath, max_sent_len):
    with open(fpath) as f:
        sent = []
        for line in f:
            line = line.strip()
            if sent and not line or len(sent) >= max_sent_len:
                yield sent
                sent = []
            w, *_ = line.split()
            sent.append(w)
        if sent:
            yield sent


def get_sentences_line(fpath, max_sent_len):
    with open(fpath) as f:
        sent = []
        for line in f:
            line = line.strip().split()
            if not line and sent:
                yield sent
                sent = []
            for w in line:
                if len(sent) >= max_sent_len:
                    yield sent
                    sent = []
                sent.append(w)
            if sent:
                yield sent
                sent = []
        if sent:
            yield sent            


def get_sentences(fpath, max_sent_len, vrt):
    if vrt:
        yield from get_sentences_vrt(fpath, max_sent_len)
    else:
        yield from get_sentences_line(fpath, max_sent_len)


def lines_from_file(fpath, tokenize=False, max_sent_len=35, vrt=False):
    """
    tokenize : bool, whether to use simple_tokenizer
    max_sent_len : int, only applicable if tokenize is False
    """
    if tokenize:
        # ignore vrt
        with open(fpath) as f:
            for sent in simple_tokenizer(line):
                yield sent, len(sent)
    else:
        for sent in get_sentences(fpath, max_sent_len, vrt):
            yield sent, len(sent)


class Tagger():
    def __init__(self, device='cpu', batch_size=100,
                 lower=False, tokenize=False, vrt=False, max_sent_len=35):
        self.device = device
        self.batch_size = batch_size
        self.lower = lower
        self.tokenize = tokenize
        self.vrt = vrt
        self.max_sent_len = max_sent_len
        self.models = []

    def add_model(self, model_path, *tasks):
        model = BaseModel.load(model_path)
        for task in tasks:
            if task not in model.label_encoder.tasks:
                raise ValueError("Model [{}] doesn't have task: {}".format(
                    model_path, task))

        self.models.append((model, tasks))

    def tag(self, sents, lengths, **kwargs):
        # lower if needed
        batch_sents = sents
        if self.lower:
            batch_sents = [[w.lower() for w in sent] for sent in sents]
        # add dummy input tasks (None)
        batch = list(zip(batch_sents, itertools.repeat(None)))
        # [token1, token2, ...]
        tokens = [token for sent in sents for token in sent]
        # output
        output = {}
        for model, tasks in self.models:
            model.to(self.device)

            inp, _ = pack_batch(model.label_encoder, batch, self.device)

            # inference
            preds = model.predict(inp, *tasks, **kwargs)

            for task in preds:
                # flatten all sentences (since some tasks return flattened output)
                if model.label_encoder.tasks[task].level != 'char':
                    preds[task] = [hyp for sent in preds[task] for hyp in sent]
                # postprocess if needed
                if model.label_encoder.tasks[task].preprocessor_fn is not None:
                    post = []
                    assert len(tokens) == len(preds[task])
                    for tok, hyp in zip(tokens, preds[task]):
                        pred = model.label_encoder.tasks[task].preprocessor_fn \
                               .inverse_transform(hyp, tok)
                        post.append(pred)
                    preds[task] = post
                # append
                output[task] = preds[task]
            model.to('cpu')

        tasks = sorted(output)

        # [(task1, task2, ...), (task1, task2, ...), ...]
        output = list(zip(*tuple(output[task] for task in tasks)))
        assert len(tokens) == len(output), "{} != {}".format(len(tokens), len(output))

        # segment sentences
        tagged = []
        for length in lengths:
            sent = []
            for _ in range(length):
                sent.append((tokens.pop(0), output.pop(0)))
            tagged.append(sent)

        return tagged, tasks

    def tag_file(self, fpath, sep='\t', keep_boundaries=False, **kwargs):
        _, ext = os.path.splitext(fpath)
        header = False

        with open(utils.ensure_ext(fpath, ext, 'pie'), 'w+') as f:
            lines = lines_from_file(
                fpath, tokenize=self.tokenize,
                max_sent_len=self.max_sent_len, vrt=self.vrt)
            lines = list(lines)

            with tqdm.tqdm(total=len(lines)) as pbar:
                for chunk in utils.chunks(lines, self.batch_size):

                    tagged, tasks = self.tag(*zip(*chunk), **kwargs)
                    pbar.update(n=len(tagged))

                    for sent in tagged:
                        if not header:
                            f.write(sep.join(['token'] + tasks) + '\n')
                            header = True
                        for token, tags in sent:
                            f.write(sep.join([token] + list(tags)) + '\n')
                        if keep_boundaries:
                            f.write('\n')




from .base_model import BaseModel
from .model import SimpleModel
from .encoder import RNNEncoder
from .embedding import CNNEmbedding, RNNEmbedding, EmbeddingConcat, EmbeddingMixer
from .embedding import build_embeddings
from .decoder import LinearDecoder, AttentionalDecoder, CRFDecoder
from .loaders import get_pretrained_embeddings
from .scorer import Scorer, compute_scores




import warnings
from functools import partial
import tarfile
import json
import yaml
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
import logging
from collections import Counter, defaultdict
import random

import torch

from pie import utils, torch_utils, constants
from . import preprocessors


class LabelEncoder(object):
    """
    Label encoder
    """
    def __init__(self, level='token', name=None, target=None,
                 lower=False, utfnorm=False, utfnorm_type='NFKD', drop_diacritics=False,
                 preprocessor=None, max_size=None, min_freq=1,
                 pad=True, eos=False, bos=False, reserved=(), **meta):

        if level.lower() not in ('token', 'char'):
            raise ValueError("`level` must be 'token' or 'char'. Got ", level)

        self.meta = meta  # dictionary with other task-relevant information
        self.pad = constants.PAD if pad else None
        self.eos = constants.EOS if eos else None
        self.bos = constants.BOS if bos else None
        self.lower = lower
        self.utfnorm = utfnorm
        self.utfnorm_type = utfnorm_type
        self.drop_diacritics = drop_diacritics
        self.text_preprocess_fn = None
        if lower or utfnorm or drop_diacritics:
            self.text_preprocess_fn = self._get_text_preprocess_fn(
                lower, utfnorm, utfnorm_type, drop_diacritics)
        self.preprocessor = preprocessor
        self.preprocessor_fn = \
            getattr(preprocessors, preprocessor) if preprocessor else None
        self.max_size = max_size
        self.min_freq = min_freq
        self.level = level.lower()
        self.target = target
        self.name = name
        self.reserved = reserved + (constants.UNK,)  # always use <unk>
        self.reserved += tuple([sym for sym in [self.eos, self.pad, self.bos] if sym])
        self.freqs = Counter()
        self.known_tokens = set()  # for char-level dicts, keep word-level known tokens
        self.table = None
        self.inverse_table = None
        self.fitted = False

    def _get_text_preprocess_fn(self, lower, utfnorm, utfnorm_type, drop_diacritics):
        fns = []
        if lower:
            fns.append(utils.lower_str)
        if utfnorm:
            fns.append(partial(utils.apply_utfnorm, form=utfnorm_type))
        if drop_diacritics:
            fns.append(utils.drop_diacritics)

        return utils.compose(*fns)

    def __len__(self):
        if not self.fitted:
            raise ValueError("Cannot get length of unfitted LabelEncoder")

        return len(self.table)

    def __eq__(self, other):
        if type(other) != LabelEncoder:
            return False

        return self.pad == other.pad and \
            self.eos == other.eos and \
            self.bos == other.bos and \
            self.preprocessor == other.preprocessor and \
            self.max_size == other.max_size and \
            self.level == other.level and \
            self.lower == other.lower and \
            self.utfnorm == other.utfnorm and \
            self.drop_diacritics == other.drop_diacritics and \
            self.target == other.target and \
            self.freqs == other.freqs and \
            self.table == other.table and \
            self.inverse_table == other.inverse_table and \
            self.fitted == other.fitted

    def __repr__(self):
        try:
            length = len(self)
        except Exception:
            length = 0

        return (
            '<LabelEncoder name="{}" lower="{}" utfnorm="{}" utfnorm_type="{}" ' +
            'target="{}" vocab="{}" level="{}" fitted="{}"/>'
        ).format(
            self.name, self.lower, self.utfnorm, self.utfnorm_type,
            self.target, self.level, length, self.fitted)

    def get_type_stats(self):
        """
        Compute number of known types, total number of types and ratio
        """
        if not self.fitted:
            raise ValueError("Vocabulary hasn't been computed yet")

        total_types = len(self.freqs)
        known_types = len(self) - len(self.reserved)
        return known_types, total_types, known_types / total_types

    def get_token_stats(self):
        """
        Compute number of known tokens, total number of types and ratio
        """
        if not self.fitted:
            raise ValueError("Vocabulary hasn't been computed yet")

        total_tokens = sum(self.freqs.values())
        known_tokens = sum(self.freqs[w] for w in self.table)
        return known_tokens, total_tokens, known_tokens / total_tokens

    def add(self, seq, rseq=None):
        if self.fitted:
            raise ValueError("Already fitted")

        postseq = self.preprocess(seq, rseq)

        if self.level == 'token':
            self.freqs.update(postseq)
        else:
            self.freqs.update(c for tok in postseq for c in tok)
            # always use original sequence for known tokens
            self.known_tokens.update(seq)

    def compute_vocab(self):
        if self.fitted:
            raise ValueError("Cannot compute vocabulary, already fitted")

        if len(self.freqs) == 0:
            logging.warning("Computing vocabulary for empty encoder {}"
                            .format(self.name))

        if self.max_size:
            most_common = self.freqs.most_common(n=self.max_size)
        elif self.min_freq:
            most_common = [it for it in self.freqs.items() if it[1] >= self.min_freq]
        else:
            most_common = self.freqs.most_common()

        self.inverse_table = list(self.reserved) + [sym for sym, _ in most_common]
        self.table = {sym: idx for idx, sym in enumerate(self.inverse_table)}
        self.fitted = True

    def preprocess_text(self, seq):
        """
        Apply surface level preprocessing such as lowering, unicode normalization
        """
        if self.text_preprocess_fn:
            seq = list(map(self.text_preprocess_fn, seq))
        return seq

    def preprocess(self, tseq, rseq=None):
        """
        Full preprocessing pipeline including possible token-level transformations
        """
        tseq = self.preprocess_text(tseq)

        if self.preprocessor_fn is not None:
            if rseq is None:
                raise ValueError("Expected ref sequence for preprocessor")

            return [self.preprocessor_fn.transform(t, r) for t, r in zip(tseq, rseq)]

        return tseq

    def transform(self, seq, rseq=None):
        if not self.fitted:
            raise ValueError("Vocabulary hasn't been computed yet")

        def transform_seq(s):
            output = []
            if self.bos:
                output.append(self.get_bos())

            for tok in s:
                output.append(self.table.get(tok, self.table[constants.UNK]))

            if self.eos:
                output.append(self.get_eos())

            return output

        # preprocess
        seq = self.preprocess(seq, rseq)

        if self.level == 'token':
            output = transform_seq(seq)
        else:
            output = [transform_seq(w) for w in seq]

        return output

    def inverse_transform(self, seq):
        if not self.fitted:
            raise ValueError("Vocabulary hasn't been computed yet")

        return [self.inverse_table[i] for i in seq]

    def stringify(self, seq, length=None):
        if not self.fitted:
            raise ValueError("Vocabulary hasn't been computed yet")

        eos, bos = self.get_eos(), self.get_bos()
        if length is not None:
            if eos is not None or bos is not None:
                warnings.warn("Length was passed to stringify but LabelEncoder "
                              "has <eos> and/or <bos> tokens")
            seq = seq[:length]
        else:
            if eos is None:
                raise ValueError("Don't know how to compute input length")
            try:
                # some generations might fail to produce the <eos> symbol
                seq = seq[:seq.index(eos)]
            except ValueError:
                pass

            # eventually remove <bos> if required
            if bos is not None:
                if len(seq) > 0 and seq[0] == bos:
                    seq = seq[1:]

        return self.inverse_transform(seq)

    def _get_sym(self, sym):
        if not self.fitted:
            raise ValueError("Vocabulary hasn't been computed yet")

        return self.table.get(sym)

    def get_pad(self):
        return self._get_sym(constants.PAD)

    def get_eos(self):
        return self._get_sym(constants.EOS)

    def get_bos(self):
        return self._get_sym(constants.BOS)

    def get_unk(self):
        return self._get_sym(constants.UNK)

    def jsonify(self):
        if not self.fitted:
            raise ValueError("Attempted to serialize unfitted encoder")

        return {'name': self.name,
                'eos': self.eos,
                'bos': self.bos,
                'pad': self.pad,
                'meta': self.meta,
                'level': self.level,
                'preprocessor': self.preprocessor,
                'lower': self.lower,
                'utfnorm': self.utfnorm,
                'utfnorm_type': self.utfnorm_type,
                'drop_diacritics': self.drop_diacritics,
                'target': self.target,
                'max_size': self.max_size,
                'min_freq': self.min_freq,
                'freqs': dict(self.freqs),
                'table': dict(self.table),
                'inverse_table': self.inverse_table,
                'known_tokens': list(self.known_tokens)}

    @classmethod
    def from_json(cls, obj):
        inst = cls(pad=obj['pad'], eos=obj['eos'], bos=obj['bos'],
                   level=obj['level'], target=obj['target'], lower=obj['lower'],
                   max_size=obj['max_size'], min_freq=['min_freq'],
                   drop_diacritics=obj.get('drop_diacritics', False),
                   utfnorm=obj.get('utfnorm', False),
                   utfnorm_type=obj.get('utfnorm_type', False),
                   preprocessor=obj.get('preprocessor'),
                   name=obj['name'], meta=obj.get('meta', {}))
        inst.freqs = Counter(obj['freqs'])
        inst.table = dict(obj['table'])
        inst.inverse_table = list(obj['inverse_table'])
        inst.known_tokens = set(obj['known_tokens'])
        inst.fitted = True

        return inst


class MultiLabelEncoder(object):
    """
    Complex Label encoder for all tasks.
    """
    def __init__(self, word_max_size=None, word_min_freq=1, word_lower=False,
                 char_max_size=None, char_min_freq=None, char_lower=False,
                 char_eos=True, char_bos=True, utfnorm=False, utfnorm_type='NFKD',
                 drop_diacritics=False):
        self.word = LabelEncoder(max_size=word_max_size, min_freq=word_min_freq,
                                 lower=word_lower, utfnorm=utfnorm,
                                 utfnorm_type=utfnorm_type,
                                 drop_diacritics=drop_diacritics, name='word')
        self.char = LabelEncoder(max_size=char_max_size, min_freq=char_min_freq,
                                 level='char', lower=char_lower, name='char',
                                 eos=char_eos, bos=char_bos, utfnorm_type=utfnorm_type,
                                 utfnorm=utfnorm, drop_diacritics=drop_diacritics)
        self.tasks = {}
        self.nsents = None

    def __repr__(self):
        return (
            '<MultiLabelEncoder>\n\t' +
            '\n\t'.join(map(str, [self.word, self.char] + list(self.tasks.values()))) +
            '\n</MultiLabelEncoder>')

    def __eq__(self, other):
        if not (self.word == other.word and self.char == other.char):
            return False

        for task in self.tasks:
            if task not in other.tasks:
                return False
            if self.tasks[task] != other.tasks[task]:
                return False

        return True

    def add_task(self, name, **meta):
        self.tasks[name] = LabelEncoder(name=name, **meta)

        # check <eos> <bos> (not suitable for linear models)
        if meta['level'].lower() != 'char' and (meta.get('eos') or meta.get('bos')):
            raise ValueError(
                ('[Task: {task}] => `bos` and `eos` options are '
                 'only compatible with char-level tasks but got '
                 'level: "{level}". Aborting!!!').format(
                    task=name, level=meta['level']))

        return self

    @classmethod
    def from_settings(cls, settings, tasks=None):
        le = cls(word_max_size=settings.word_max_size,
                 word_min_freq=settings.word_min_freq,
                 word_lower=settings.word_lower,
                 char_max_size=settings.char_max_size,
                 char_min_freq=settings.char_min_freq,
                 char_lower=settings.char_lower,
                 char_eos=settings.char_eos,
                 char_bos=settings.char_bos,
                 utfnorm=settings.utfnorm,
                 utfnorm_type=settings.utfnorm_type,
                 drop_diacritics=settings.drop_diacritics)

        for task in settings.tasks:
            if tasks is not None and task['settings']['target'] not in tasks:
                raise ValueError("No available data for task [{}]".format(
                    task['settings']['target']))
            le.add_task(task['name'], level=task['level'], **task['settings'])

        return le

    def fit(self, lines):
        """
        Parameters
        ===========
        lines : iterator over tuples of (Input, Tasks)
        """
        for idx, inp in enumerate(lines):
            tasks = None
            if isinstance(inp, tuple):
                inp, tasks = inp

            # input
            self.word.add(inp)
            self.char.add(inp)

            for le in self.tasks.values():
                le.add(tasks[le.target], inp)

        self.word.compute_vocab()
        self.char.compute_vocab()
        for le in self.tasks.values():
            le.compute_vocab()

        return self

    def fit_reader(self, reader):
        """
        fit reader in a non verbose way (to warn about parsing issues)
        """
        return self.fit(line for (_, line) in reader.readsents(silent=False))

    def transform(self, sents):
        """
        Parameters
        ===========
        sents : list of Example's

        Returns
        ===========
        tuple of (word, char), task_dict

            - word: list of integers
            - char: list of integers where each list represents a word at the
                character level
            - task_dict: Dict to corresponding integer output for each task
        """
        word, char, tasks_dict = [], [], defaultdict(list)

        for inp in sents:
            tasks = None

            # task might not be passed
            if isinstance(inp, tuple):
                inp, tasks = inp

            # input data
            word.append(self.word.transform(inp))
            char.extend(self.char.transform(inp))

            # task data
            if tasks is None:
                # during inference there is no task data (pass None)
                continue

            for le in self.tasks.values():
                task_data = le.transform(tasks[le.target], inp)
                # add data
                if le.level == 'char':
                    tasks_dict[le.name].extend(task_data)
                else:
                    tasks_dict[le.name].append(task_data)

        return (word, char), tasks_dict

    def jsonify(self):
        return {'word': self.word.jsonify(),
                'char': self.char.jsonify(),
                'tasks': {le.name: le.jsonify() for le in self.tasks.values()}}

    def save(self, path):
        with open(path, 'w+') as f:
            yaml.dump(self.jsonify(), f, Dumper=Dumper)

    @staticmethod
    def _init(inst, obj):
        inst.word = LabelEncoder.from_json(obj['word'])
        inst.char = LabelEncoder.from_json(obj['char'])

        for task, le in obj['tasks'].items():
            inst.tasks[task] = LabelEncoder.from_json(le)

        return inst

    @classmethod
    def load_from_string(cls, string):
        inst = cls()
        try:
            obj = json.loads(string)
        except ValueError:      # use yaml
            obj = yaml.load(string, Loader=Loader)
        return cls._init(inst, obj)

    @classmethod
    def load_from_file(cls, path):
        with open(path, 'r+') as f:
            return cls.load_from_string(f.read())

    @classmethod
    def load_from_pretrained_model(cls, path):
        with tarfile.open(utils.ensure_ext(path, 'tar'), 'r') as tar:
            return cls.load_from_string(utils.get_gzip_from_tar(tar, 'label_encoder'))


class Dataset(object):
    """
    Dataset class to encode files into integers and compute batches.

    Settings
    ===========
    buffer_size : int, maximum number of sentences in memory at any given time.
       The larger the buffer size the more memory instensive the dataset will
       be but also the more effective the shuffling over instances.
    batch_size : int, number of sentences per batch
    device : str, target device to put the processed batches on
    shuffle : bool, whether to shuffle items in the buffer
    minimize_pad : bool, whether to pack batches with sentences of similar length
       in order to minimize padding.

    Parameters
    ===========
    input_path : str (optional), either a path to a directory, a path to a file
        or a unix style pathname pattern expansion for glob. If given, the
        input_path in settings will be overwritten by the new value
    label_encoder : optional, prefitted LabelEncoder object
    """
    def __init__(self, settings, reader, label_encoder):

        if settings.batch_size > settings.buffer_size:
            raise ValueError("Not enough buffer capacity {} for batch_size of {}"
                             .format(settings.buffer_size, settings.batch_size))

        # attributes
        self.buffer_size = settings.buffer_size
        self.batch_size = settings.batch_size
        self.device = settings.device
        self.shuffle = settings.shuffle
        self.minimize_pad = settings.minimize_pad
        self.cache_dataset = settings.cache_dataset

        # data
        self.reader = reader
        self.label_encoder = label_encoder
        self.cached = []

    @staticmethod
    def get_nelement(batch):
        """
        Returns the number of elements in a batch (based on word-level length)
        """
        return batch[0][0][1].sum().item()

    def pack_batch(self, batch, device=None):
        """
        Transform batch data to tensors
        """
        return pack_batch(self.label_encoder, batch, device or self.device)

    def prepare_buffer(self, buf, return_raw=False, **kwargs):
        "Transform buffer into batch generator"

        def key(data):
            inp, tasks = data
            return len(inp)

        if self.minimize_pad:
            buf = sorted(buf, key=key, reverse=True)
        elif self.shuffle:
            random.shuffle(buf)

        batches = list(utils.chunks(buf, self.batch_size))

        if self.shuffle:
            random.shuffle(batches)

        for batch in batches:
            packed = self.pack_batch(batch, **kwargs)
            if return_raw:
                inp, tasks = zip(*batch)
                yield packed, (inp, tasks)
            else:
                yield packed

    def batch_generator(self, return_raw=False):
        """
        Generator over dataset batches. Each batch is a tuple of (input, tasks):
            * (word, char)
                - word : tensor(length, batch_size), padded lengths
                - char : tensor(length, batch_size * words), padded lengths
            * (tasks) dictionary with tasks
        """
        if self.cache_dataset:
            if not self.cached:
                self.cache_batches()
            if self.shuffle:
                random.shuffle(self.cached)

            for batch, raw in self.cached:
                # move to device
                batch = tuple(list(wrap_device(batch, self.device)))
                if return_raw:
                    yield batch, raw
                else:
                    yield batch
        else:
            yield from self.batch_generator_(return_raw=return_raw)

    def batch_generator_(self, return_raw=False):
        buf = []
        for (fpath, line_num), data in self.reader.readsents():

            # fill buffer
            buf.append(data)

            # check if buffer is full and yield
            if len(buf) == self.buffer_size:
                yield from self.prepare_buffer(buf, return_raw=return_raw)
                buf = []

        if len(buf) > 0:
            yield from self.prepare_buffer(buf, return_raw=return_raw)

    def cache_batches(self):
        if self.cached:
            return

        buf = [data for _, data in self.reader.readsents()]
        for batch, raw in self.prepare_buffer(buf, return_raw=True, device='cpu'):
            self.cached.append((batch, raw))


def pack_batch(label_encoder, batch, device=None):
    """
    Transform batch data to tensors
    """
    (word, char), tasks = label_encoder.transform(batch)

    word = torch_utils.pad_batch(word, label_encoder.word.get_pad(), device=device)
    char = torch_utils.pad_batch(char, label_encoder.char.get_pad(), device=device)

    output_tasks = {}
    for task, data in tasks.items():
        output_tasks[task] = torch_utils.pad_batch(
            data, label_encoder.tasks[task].get_pad(), device=device)

    return (word, char), output_tasks


def wrap_device(it, device):
    for i in it:
        if isinstance(i, torch.Tensor):
            yield i.to(device)
        elif isinstance(i, dict):
            yield {k: tuple(wrap_device(v, device)) for k, v in i.items()}
        else:
            yield tuple(wrap_device(i, device))


class device_wrapper(object):
    def __init__(self, batches, device):
        self.batches = batches
        self.device = device

    def __getitem__(self, idx):
        return tuple(wrap_device(self.batches[idx], self.device))

    def __len__(self):
        return len(self.batches)



"""
    For more information about how to run the webapp, go check app.py in
    the root directory.
    This module is available so that you can enhance or customize your app
    by changing its tokenizer for example

"""
from os import getenv
from typing import Iterator, List, Tuple, Callable, Iterable
from flask import Flask, request, Response, stream_with_context, current_app

from pie.tagger import Tagger, simple_tokenizer
from pie.utils import chunks, model_spec

MODEL_FILE = getenv("PIE_MODEL")
BATCH = int(getenv("PIE_BATCH", 3))
DEVICE = getenv("PIE_DEVICE", "cpu")


Tokenizer = Callable[[str, bool], Iterable[List[str]]]


class DataIterator:
    def __init__(self, tokenizer: Tokenizer = None):
        """ Iterator used to parse the text and returns bits to tag

        :param tokenizer: Tokenizer
        """
        self.tokenizer = tokenizer or simple_tokenizer

    def __call__(self, data: str, lower: bool = False) -> Iterable[Tuple[List[str], int]]:
        """ Default iter data takes a text, an option to make lower
        and yield lists of words along with the length of the list

        :param data: A plain text
        :param lower: Whether or not to lower the text
        :yields: (Sentence as a list of word, Size of the sentence)
        """
        for sentence in self.tokenizer(data, lower=lower):
            yield sentence, len(sentence)


class Formatter:
    def __init__(self, tasks: List[str]):
        self.tasks = tasks

    def format_headers(self)-> List[str]:
        """ Format the headers """
        return ["token"] + self.tasks

    def format_line(self, token: str, tags: Iterable[str]) -> List[str]:
        """ Format the tags"""
        return [token] + list(tags)


def bind(app: Flask = None, device: str = None, batch_size: int = None,
         model_file: str = None, formatter_class: Formatter = None,
         tokenizer: Tokenizer = None, allow_origin: str = None,
         route_path: str= "/", headers=None) -> Flask:
    """ Binds default value

    :param app: Flask app to bind with the tagger
    :param device: Device to use for PyTorch (Default : cpu)
    :param batch_size: Size of the batch to treat
    :param model_file: Model to use to tag
    :param formatter_class: Formatter of response
    :param tokenizer: Tokenizer to split text into segments (eg. sentence) and into words
    :param allow_origin: Value for the http header field Access-Control-Allow-Origin
    :param route_path: Route for the API (default : `/`)
    :param headers: Additional headers
    :returns: Application
    """
    # Generates or use default values for non completed parameters
    if not app:
        app = Flask(__name__)

    if not batch_size:
        batch_size = BATCH

    device = device or DEVICE
    model_file = model_file or MODEL_FILE
    allow_origin = allow_origin or '*'

    if tokenizer:
        data_iterator = DataIterator(tokenizer)
    else:
        data_iterator = DataIterator()

    tagger = Tagger(device=device, batch_size=batch_size)

    for model, tasks in model_spec(model_file):
        tagger.add_model(model, *tasks)

    formatter_class = formatter_class or Formatter

    _headers = {
        'Content-Type': 'text/plain; charset=utf-8',
        'Access-Control-Allow-Origin': allow_origin
       }
    if headers:
        _headers.update(headers)

    @app.route(route_path, methods=["POST", "GET", "OPTIONS"])
    def lemmatize():
        def lemmatization_stream() -> Iterator[str]:
            lower = request.args.get("lower", False)
            if lower:
                lower = True

            if request.method == "GET":
                data = request.args.get("data")
            else:
                data = request.form.get("data")

            if not data:
                yield ""
            else:
                header = False
                for chunk in chunks(data_iterator(data, lower=lower), size=BATCH):
                    sents, lengths = zip(*chunk)

                    tagged, tasks = tagger.tag(sents=sents, lengths=lengths)
                    formatter = formatter_class(tasks)
                    sep = "\t"
                    for sent in tagged:
                        if not header:
                            yield sep.join(formatter.format_headers()) + '\r\n'
                            header = True
                        for token, tags in sent:
                            yield sep.join(formatter.format_line(token, tags)) + '\r\n'

        return Response(
               stream_with_context(lemmatization_stream()),
               200,
               headers=_headers
        )

    return app




import os
import unittest

from pie.data import Dataset, Reader, MultiLabelEncoder
from pie.settings import settings_from_file


testpath = os.path.join(os.path.dirname(__file__), 'testconfig.json')


class TestLabelEncoderSerialization(unittest.TestCase):
    def setUp(self):
        settings = settings_from_file(testpath)
        reader = Reader(settings, settings.input_path)
        label_encoder = MultiLabelEncoder.from_settings(settings)
        label_encoder.fit_reader(reader)
        self.data = Dataset(settings, reader, label_encoder)

    def test_serialization(self):
        le = self.data.label_encoder
        le.save('/tmp/encoder.json')
        le2 = MultiLabelEncoder.load_from_file('/tmp/encoder.json')

        self.assertEqual(len(le.tasks), len(le2.tasks),
                         "Unequal number of Modality encoders")

        self.assertEqual(le.word, le2.word, "word encoder")
        self.assertEqual(le.char, le2.char, "char encoder")

        for task in le.tasks:
            self.assertTrue(
                le.tasks[task] == le2.tasks[task],
                "Unequal serialized label encoder for task {}".format(task))


class TestWordCharEncoding(unittest.TestCase):
    def setUp(self):
        settings = settings_from_file(testpath)
        reader = Reader(settings, settings.input_path)
        label_encoder = MultiLabelEncoder.from_settings(settings)
        label_encoder.fit_reader(reader)
        self.data = Dataset(settings, reader, label_encoder)

    def test_lengths(self):
        ((word, wlen), (char, clen)), _ = next(self.data.batch_generator())

        for c, cl in zip(char.t(), clen):
            self.assertEqual(c[0].item(), self.data.label_encoder.char.get_bos())
            self.assertEqual(c[cl-1].item(), self.data.label_encoder.char.get_eos())

    def test_word_char(self):
        for ((word, wlen), (char, clen)), _ in self.data.batch_generator():
            idx = 0
            total_words = 0
            for sent, nwords in zip(word.t(), wlen):
                for word in sent[:nwords]:
                    # get word
                    word = self.data.label_encoder.word.inverse_table[word]
                    # get chars
                    chars = char.t()[idx][1:clen[idx]-1].tolist()  # remove <eos>,<bos>
                    chars = ''.join(self.data.label_encoder.char.inverse_transform(chars))
                    self.assertEqual(word, chars)
                    idx += 1
                total_words += nwords
            self.assertEqual(idx, total_words, "Checked all words")


def _test_conversion(settings, level='token'):
    reader = Reader(settings, settings.input_path)
    label_encoder = MultiLabelEncoder.from_settings(settings)
    label_encoder.fit_reader(reader)
    data = Dataset(settings, reader, label_encoder)

    le = label_encoder.tasks['lemma']
    for (inp, tasks), (rinp, rtasks) in data.batch_generator(return_raw=True):
        # preds
        tinp, tlen = tasks['lemma']
        preds = [le.stringify(t, l) for t, l in zip(tinp.t().tolist(), tlen.tolist())]
        if level == 'token':
            preds = [w for line in preds for w in line]
        # tokens
        tokens = [tok for line in rinp for tok in line]
        # trues
        trues = [w for line in rtasks for w in line['lemma']]

        # check
        for pred, token, true in zip(preds, tokens, trues):
            rec = le.preprocessor_fn.inverse_transform(pred, token)
            assert rec == true, (pred, token, true, rec)


class TestGreedyScripts(unittest.TestCase):
    def setUp(self):
        settings = settings_from_file(testpath)
        settings.tasks = [
            {"name": "lemma", "level": "char", "decoder": "linear",
             "settings": {"preprocessor": "greedy_scripts", "target": "lemma"}}
        ]
        settings.char_eos = False
        settings.char_bos = False
        settings.char_max_size = 10000
        self.settings = settings

    def testConversion(self):
        _test_conversion(self.settings, level="char")


class TestEditTrees(unittest.TestCase):
    def setUp(self):
        settings = settings_from_file(testpath)
        settings.tasks = [
            {"name": "lemma", "level": "token", "decoder": "linear",
             "settings": {"preprocessor": "edit_trees", "target": "lemma"}}
        ]
        self.settings = settings

    def testConversion(self):
        _test_conversion(self.settings, level="token")


class TestEditTreeTuples(unittest.TestCase):
    def setUp(self):
        settings = settings_from_file(testpath)
        settings.tasks = [
            {"name": "lemma", "level": "char", "decoder": "attentional",
             "settings": {"preprocessor": "edit_tree_tuples", "target": "lemma"}}
        ]
        self.settings = settings

    def testConversion(self):
        _test_conversion(self.settings, level="char")




import torch
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence


def init_hidden_for(inp, num_dirs, num_layers, hid_dim, cell,
                    h_0=None, add_init_jitter=False):
    """
    General function for initializing RNN hidden states

    Parameters:
    - inp: torch.Tensor(seq_len, batch_size, dim)
    """
    size = (num_dirs * num_layers, inp.size(1), hid_dim)

    # create h_0
    if h_0 is not None:
        h_0 = h_0.repeat(1, inp.size(1), 1)
    else:
        h_0 = torch.zeros(*size, device=inp.device)

    # eventualy add jitter
    if add_init_jitter:
        h_0 = h_0 + torch.normal(torch.zeros_like(h_0), 0.3)

    if cell.startswith('LSTM'):
        # compute memory cell
        return h_0, torch.zeros_like(h_0)
    else:
        return h_0


def pack_sort(inp, lengths, batch_first=False):
    """
    Transform input into PaddedSequence sorting batch by length (as required).
    Also return an index variable that unsorts the output back to the original
    order.

    Parameters:
    -----------
    inp: torch.Tensor(seq_len x batch x dim)
    lengths: LongTensor of length ``batch``

    >>> from torch.nn.utils.rnn import pad_packed_sequence as unpack
    >>> inp = torch.tensor([[1, 3], [2, 4], [0, 5]], dtype=torch.float)
    >>> lengths = torch.tensor([2, 3]) # unsorted order
    >>> sorted_inp, unsort = pack_sort(inp, lengths)
    >>> sorted_inp, _ = unpack(sorted_inp)
    >>> sorted_inp[:, unsort].tolist()  # original order
    [[1.0, 3.0], [2.0, 4.0], [0.0, 5.0]]
    >>> sorted_inp.tolist()  # sorted by length
    [[3.0, 1.0], [4.0, 2.0], [5.0, 0.0]]
    """
    if not isinstance(lengths, torch.Tensor):
        lengths = torch.tensor(lengths)  # no need to use gpu

    lengths, sort = torch.sort(lengths, descending=True)
    _, unsort = sort.sort()

    if batch_first:
        inp = pack_padded_sequence(inp[sort], lengths.tolist())
    else:
        inp = pack_padded_sequence(inp[:, sort], lengths.tolist())

    return inp, unsort


def get_last_token(t, lengths):
    """
    Grab last hidden activation of each batch element according to `lengths`

    #                               ^ (1)      ^ (2)      ^ (3)
    >>> t = torch.tensor([[[1],[2],[3]], [[2],[3],[4]], [[3],[4],[5]]])
    >>> lengths = torch.tensor([3, 2, 1])
    >>> get_last_token(t, lengths).tolist()
    [[3], [3], [3]]
    """
    if not isinstance(lengths, torch.Tensor):
        lengths = torch.tensor(lengths)

    seq_len, batch, _ = t.size()
    index = torch.arange(0, batch, dtype=torch.int64, device=t.device) * seq_len
    index = index + (lengths - 1)
    t = t.transpose(0, 1).contiguous()  # make it batch first
    t = t.view(seq_len * batch, -1)
    t = t.index_select(0, index)
    t = t.view(batch, -1)
    return t


def pad_flat_batch(emb, nwords, maxlen=None):
    """
    Transform a 2D flat batch (batch of words in multiple sentences) into a 3D
    padded batch where words have been allocated to their respective sentence
    according to user passed sentence lengths `nwords`

    Parameters
    ===========
    emb : torch.Tensor(total_words x emb_dim), flattened tensor of word embeddings
    nwords : torch.Tensor(batch), number of words per sentence

    Returns
    =======
    torch.Tensor(max_seq_len x batch x emb_dim) where:
        - max_seq_len = max(nwords)
        - batch = len(nwords)

    >>> emb = [[0], [1], [2], [3], [4], [5]]
    >>> nwords = [3, 1, 2]
    >>> pad_flat_batch(torch.tensor(emb), torch.tensor(nwords)).tolist()
    [[[0], [3], [4]], [[1], [0], [5]], [[2], [0], [0]]]
    """
    if len(emb) != sum(nwords):
        raise ValueError("Got {} items but was asked to pad {}"
                         .format(len(emb), sum(nwords).item()))

    output, last = [], 0
    maxlen = maxlen or max(nwords).item()

    for sentlen in nwords.tolist():
        padding = (0, 0, 0, maxlen - sentlen)
        output.append(F.pad(emb[last:last+sentlen], padding))
        last = last + sentlen

    # (seq_len x batch x emb_dim)
    output = torch.stack(output, dim=1)

    return output


def flatten_padded_batch(batch, nwords):
    """
    Inverse of pad_flat_batch

    Parameters
    ===========
    batch : tensor(seq_len, batch, encoding_size), output of the encoder
    nwords : tensor(batch), lengths of the sequence (without padding)

    Returns
    ========
    tensor(nwords, encoding_size)

    >>> batch = [[[0], [3], [4]], [[1], [0], [5]], [[2], [0], [0]]]
    >>> nwords = [3, 1, 2]
    >>> flatten_padded_batch(torch.tensor(batch), torch.tensor(nwords)).tolist()
    [[0], [1], [2], [3], [4], [5]]
    """
    with torch.no_grad():
        output = []
        for sent, sentlen in zip(batch.transpose(0, 1), nwords):
            output.extend(list(sent[:sentlen].chunk(sentlen)))  # remove <eos>

        return torch.cat(output, dim=0)


def pad_batch(batch, padding_id, device='cpu', return_lengths=True):
    """
    Pad batch into tensor
    """
    lengths = [len(example) for example in batch]
    maxlen, batch_size = max(lengths), len(batch)
    output = torch.zeros(
        maxlen, batch_size, device=device, dtype=torch.int64
    ) + padding_id

    for i, example in enumerate(batch):
        output[0:lengths[i], i].copy_(
            torch.tensor(example, dtype=torch.int64, device=device))

    if return_lengths:
        lengths = torch.tensor(lengths, dtype=torch.int64, device=device)
        return output, lengths

    return output


def pad(batch, pad=0, pos='pre'):
    """
    >>> batch = torch.tensor([[1, 1], [2, 2], [3, 3], [4, 4]])
    >>> pad(batch, pad=-1, pos='pre').tolist()
    [[-1, -1], [1, 1], [2, 2], [3, 3], [4, 4]]
    >>> pad(batch, pad=5, pos='post').tolist()
    [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]
    """
    if pos.lower() == 'pre':
        padding = (0, 0) * (batch.dim() - 1) + (1, 0)
    elif pos.lower() == 'post':
        padding = (0, 0) * (batch.dim() - 1) + (0, 1)
    else:
        raise ValueError("Unknown value for pos: {}".format(pos))

    return F.pad(batch, padding, value=pad)


def make_length_mask(lengths):
    """
    Compute binary length mask.

    lengths: torch.Tensor(batch, dtype=int) should be on the desired
        output device.

    Returns
    =======

    mask: torch.ByteTensor(batch x seq_len)
    """
    maxlen, batch = lengths.detach().max(), len(lengths)
    return torch.arange(0, maxlen, dtype=torch.int64, device=lengths.device) \
                .repeat(batch, 1) \
                .lt(lengths.unsqueeze(1))


def word_dropout(inp, p, training, encoder):
    """
    Drop input words based on frequency
    """
    # don't do anything during training
    if p == 0.0 or not training:
        return inp

    # (seq_len x batch)
    seq_len, batch = inp.size()
    # get batch of word frequencies
    mask = torch.tensor(
        [[encoder.freqs[encoder.inverse_table[inp[i, j].item()]]
          for j in range(batch)] for i in range(seq_len)]
    ).float().to(inp.device)
    # compute bernoulli mask
    # PyTorch 1.2 and 1.3 :
    #   - Deprecated masked_fill_ used with int.
    #   - Mask should be booleans in general
    mask = (1 - torch.bernoulli(mask / (p + mask))).bool()
    # don't drop padding
    mask.masked_fill_(inp.eq(encoder.get_pad()).bool(), 0)
    # set words to unknowns
    return inp.masked_fill(mask, encoder.get_unk())


def sequential_dropout(inp, p, training):
    if not training or not p:
        return inp

    mask = inp.new(1, inp.size(1), inp.size(2)).bernoulli_(1 - p)
    mask = mask / (1 - p)

    return inp * mask.expand_as(inp)


def log_sum_exp(x, dim=-1):
    """
    Numerically stable log_sum_exp

    Parameters
    ==========
    x : torch.tensor

    >>> import torch
    >>> x = torch.randn(10, 5)
    """
    max_score, _ = torch.max(x, dim)
    max_score_broadcast = max_score.unsqueeze(dim).expand_as(x)
    return max_score + (x - max_score_broadcast).exp().sum(dim).log()


def viterbi_decode(tag_sequence, transition):
    """
    Perform Viterbi decoding in log space over a sequence given a transition matrix
    specifying pairwise (transition) potentials between tags and a matrix of shape
    (sequence_length, num_tags) specifying unary potentials for possible tags per
    timestep.

    Parameters
    ==========
    tag_sequence: torch.Tensor, required.
        A tensor of shape (sequence_length, num_tags) representing scores for
        a set of tags over a given sequence.
    trans: torch.Tensor, required.
        A tensor of shape (num_tags, num_tags) representing the binary potentials
        for transitioning between a given pair of tags.

    Returns
    =======
    viterbi_path: The tag indices of the maximum likelihood tag sequence
    viterbi_score: float, The score of the viterbi path
    """
    seq_len, vocab = tag_sequence.size()

    path_scores = []
    path_indices = []

    path_scores.append(tag_sequence[0, :])

    # Evaluate the scores for all possible paths.
    for t in range(1, seq_len):

        # Add pairwise potentials to current scores.
        summed_potentials = path_scores[t - 1].unsqueeze(-1) + transition
        scores, paths = torch.max(summed_potentials, 0)
        path_scores.append(tag_sequence[t, :] + scores.squeeze())
        path_indices.append(paths.squeeze())

    # Construct the most likely sequence backwards.
    viterbi_score, best_path = torch.max(path_scores[-1].cpu(), 0)
    viterbi_path = [int(best_path.numpy())]
    for backward_t in reversed(path_indices):
        viterbi_path.append(int(backward_t[viterbi_path[-1]]))

    # Reverse the backward path.
    viterbi_path.reverse()
    
    return viterbi_path, viterbi_score



import re
from .base_reader import BaseReader, LineParseException, MissingDefaultException


class LineParser(object):
    """
    Inner class to handle sentence breaks
    """
    def __init__(self, tasks, sep, breakline_ref, breakline_data, reader):
        self.reader = reader
        self.sep = sep
        # breakline info
        self.breakline_ref = breakline_ref
        self.breakline_data = re.compile(breakline_data)
        # data
        self.inp = []
        self.tasks = {task: [] for task in tasks}

    def add(self, line, linenum):
        """
        Adds line to current sentence.
        """
        inp, *tasks = line.split(self.sep)

        if len(tasks) < len(self.tasks):
            try:
                tasks = [self.reader.get_default(task, inp) for task in self.tasks]
            except MissingDefaultException:
                raise LineParseException(
                    "Not enough number of tasks. Expected {} but got {} at line {}."
                    .format(len(self.tasks), len(tasks), linenum))

        self.inp.append(inp)

        for task, data in zip(self.tasks.keys(), tasks):
            try:
                data = getattr(self, 'process_{}'.format(task.lower()))(data)
            except AttributeError:
                pass
            finally:
                self.tasks[task].append(data)

    def check_breakline(self):
        """
        Check if sentence is finished.
        """
        if len(self.inp) == 0:
            return False

        if self.breakline_ref:
            ref = self.inp
            if self.breakline_ref != 'input':
                ref = self.tasks[self.breakline_ref]
            return self.breakline_data.match(ref[-1])

    def reset(self, nitems=None):
        """
        Reset sentence data
        """
        if nitems is None:
            self.tasks, self.inp = {task: [] for task in self.tasks}, []
        else:
            self.inp = self.inp[nitems:]
            self.tasks = {task: tdata[nitems:] for task, tdata in self.tasks.items()}


class TabReader(BaseReader):
    """
    Reader for files in tab format where each line has annotations for a given word
    and each annotation is located in a column separated by tabs.

    ...
    italiae	italia	NE
    eo	is	PRO
    tasko	taskus	NN
    ...

    Settings
    ===========
    header : bool, whether file has header.
    tasks_order : list, in case of missing header, the expected order of tasks in
        the file (there might be less)
    breakline_data : str
    breakline_ref : str, tab used to decide on line breaks
    max_sent_len : int, break lines to this length if they'd become longer
    tasks[task].default : str, method to use to fill in missing values
    """
    def __init__(self, settings, fpath, line_parser=LineParser):
        self.header = settings.header  # needed for get_tasks
        self.tasks_order = settings.tasks_order
        self.sep = settings.sep
        super(TabReader, self).__init__(settings, fpath)
        self.line_parser = line_parser
        self.breakline_ref = settings.breakline_ref
        self.breakline_data = settings.breakline_data
        self.max_sent_len = settings.max_sent_len

    def parselines(self):
        """
        Generator over sentences in a single file
        """
        with open(self.fpath, 'r+') as f:
            if self.header:
                next(f)

            parser = self.line_parser(
                self.tasks, self.sep, self.breakline_ref, self.breakline_data, self)

            for line_num, line in enumerate(f):
                line = line.strip()

                # Empty line as chunk break
                if not line and len(parser.inp) > 0:
                        yield parser.inp, parser.tasks
                        parser.reset()
                        continue
                # brealine configuration as chunk break
                elif parser.check_breakline():
                    yield parser.inp, parser.tasks
                    parser.reset()
                # max size as chunk break
                elif len(parser.inp) > self.max_sent_len:
                    inp = parser.inp[:self.max_sent_len]
                    tasks = {}
                    for t in parser.tasks:
                        tasks[t] = parser.tasks[t][:self.max_sent_len]
                    yield inp, tasks
                    parser.reset(self.max_sent_len)

                if line:
                    try:
                        parser.add(line, line_num)
                    except LineParseException as e:
                        yield e

            if len(parser.inp) > 0:
                yield parser.inp, parser.tasks

    def get_tasks(self):
        """
        Guess tasks from file assuming expected order
        """
        with open(self.fpath, 'r+') as f:

            if self.header:
                # TODO: this assumes text is first field
                _, *header = next(f).strip().split(self.sep)
                return tuple(header)

            else:
                # move to first non empty line
                line = next(f).strip()
                while not line:
                    line = next(f).strip()

                _, *tasks = line.split(self.sep)
                if len(tasks) == 0:
                    raise ValueError("Not enough input tasks: [{}]".format(self.fpath))

                return tuple(self.tasks_order[:len(tasks)])




import torch
import torch.nn as nn
import torch.nn.functional as F

from pie import torch_utils
from pie import initialization

from .lstm import CustomBiLSTM
from .highway import Highway


class CNNEmbedding(nn.Module):
    """
    Character-level Embeddings with Convolutions following Kim 2014.
    """
    def __init__(self, num_embeddings, embedding_dim, padding_idx=None,
                 highway_layers=2, kernel_sizes=(5, 4, 3), out_channels=32):
        self.num_embeddings = num_embeddings
        self.embedding_dim = out_channels * len(kernel_sizes)
        self.kernel_sizes = kernel_sizes
        self.out_channels = out_channels
        super().__init__()

        self.emb = nn.Embedding(
            num_embeddings, embedding_dim, padding_idx=padding_idx)

        convs = []
        for W in kernel_sizes:
            padding = ((W//2) - 1, W - (W//2), 0, 0)
            conv = torch.nn.Sequential(
                torch.nn.ZeroPad2d(padding),
                torch.nn.Conv2d(1, out_channels, (embedding_dim, W)))
            convs.append(conv)
        self.convs = nn.ModuleList(convs)

        self.highway = None
        if highway_layers > 0:
            self.highway = Highway(self.embedding_dim, highway_layers)

        self.init()

    def init(self):
        initialization.init_embeddings(self.emb)
        for conv_seq in self.convs:
            for conv in conv_seq:
                if isinstance(conv, torch.nn.ZeroPad2d):
                    continue
            initialization.init_conv(conv)

    def forward(self, char, nchars, nwords):
        emb = self.emb(char)

        emb = emb.transpose(0, 1)  # (batch x seq_len x emb_dim)
        emb = emb.transpose(1, 2)  # (batch x emb_dim x seq_len)
        emb = emb.unsqueeze(1)     # (batch x 1 x emb_dim x seq_len)

        conv_outs = []
        for conv in self.convs:
            # (batch x C_o x seq_len)
            conv_outs.append(F.relu(conv(emb).squeeze(2)))

        # (batch * nwords x C_o * len(kernel_sizes) x seq_len)
        conv_outs = torch.cat(conv_outs, dim=1)
        # (batch * nwords  x C_o * len(kernel_sizes) x 1)
        conv_out = F.max_pool1d(conv_outs, conv_outs.size(2)).squeeze(2)
        if self.highway is not None:
            conv_out = self.highway(conv_out)
        conv_out = torch_utils.pad_flat_batch(
            conv_out, nwords, maxlen=max(nwords).item())

        conv_outs = conv_outs.transpose(0, 2).transpose(1, 2)

        return conv_out, conv_outs


class RNNEmbedding(nn.Module):
    """
    Character-level Embeddings with BiRNNs.
    """
    def __init__(self, num_embeddings, embedding_dim, padding_idx=None,
                 custom_lstm=False, cell='LSTM', init_rnn='default',
                 num_layers=1, dropout=0.0):
        self.num_embeddings = num_embeddings
        self.num_layers = num_layers
        self.embedding_dim = embedding_dim * 2  # bidirectional
        super().__init__()

        self.emb = nn.Embedding(num_embeddings, embedding_dim, padding_idx=padding_idx)
        initialization.init_embeddings(self.emb)

        if custom_lstm:
            self.rnn = CustomBiLSTM(
                embedding_dim, embedding_dim, num_layers=num_layers, dropout=dropout)
        else:
            self.rnn = getattr(nn, cell)(
                embedding_dim, embedding_dim, bidirectional=True,
                num_layers=num_layers, dropout=dropout if num_layers > 1 else 0)
            initialization.init_rnn(self.rnn, scheme=init_rnn)

    def forward(self, char, nchars, nwords):
        """
        Parameters
        ===========
        char : tensor((seq_len x) batch)
        nchars : tensor(batch)
        nwords : tensor(output batch)
        """
        # embed
        char = self.emb(char)
        # rnn
        hidden = None
        _, sort = torch.sort(nchars, descending=True)
        _, unsort = sort.sort()
        char, nchars = char[:, sort], nchars[sort]
        if isinstance(self.rnn, nn.RNNBase):
            outs, emb = self.rnn(
                nn.utils.rnn.pack_padded_sequence(char, nchars.cpu()), hidden)
            outs, _ = nn.utils.rnn.pad_packed_sequence(outs)
            if isinstance(emb, tuple):
                emb, _ = emb
        else:
            outs, (emb, _) = self.rnn(char, hidden, nchars)
        # (max_seq_len x batch * nwords x emb_dim)
        outs, emb = outs[:, unsort], emb[:, unsort]
        # (layers * 2 x batch x hidden) -> (layers x 2 x batch x hidden)
        emb = emb.view(self.num_layers, 2, len(nchars), -1)
        # use only last layer
        emb = emb[-1]
        # (2 x batch x hidden) - > (batch x 2 * hidden)
        emb = emb.transpose(0, 1).contiguous().view(len(nchars), -1)
        # (batch x 2 * hidden) -> (nwords x batch x 2 * hidden)
        emb = torch_utils.pad_flat_batch(emb, nwords, maxlen=max(nwords).item())

        return emb, outs


class EmbeddingMixer(nn.Module):
    def __init__(self, emb_dim):
        self.embedding_dim = emb_dim
        super().__init__()

        # mix parameter
        self.alpha = nn.Parameter(
            torch.Tensor(self.embedding_dim * 2, 1).uniform_(-0.05, 0.05))

    def forward(self, wembs, cembs):
        # ((seq_len x) batch x emb_dim * 2)
        alpha_in = torch.cat([wembs, cembs], dim=-1)
        # ((seq_len x) batch)
        if wembs.dim() == 3:
            alpha = torch.sigmoid(torch.einsum('do,mbd->mb', [self.alpha, alpha_in]))
        else:
            alpha = torch.sigmoid(torch.einsum('do,bd->b', [self.alpha, alpha_in]))

        wembs = alpha.unsqueeze(-1).expand_as(wembs) * wembs
        cembs = (1 - alpha).unsqueeze(-1).expand_as(cembs) * cembs

        return wembs + cembs


def EmbeddingConcat():
    def func(wemb, cemb):
        return torch.cat([wemb, cemb], dim=-1)
    return func


def build_embeddings(label_encoder, wemb_dim,
                     cemb_dim, cemb_type, custom_cemb_cell, cemb_layers, cell, init_rnn,
                     merge_type, dropout):
    """
    Utility function to build embedding layers
    """
    wemb = None
    if wemb_dim > 0:
        wemb = nn.Embedding(len(label_encoder.word), wemb_dim,
                            padding_idx=label_encoder.word.get_pad())
        # init embeddings
        initialization.init_embeddings(wemb)

    cemb = None
    if cemb_type.upper() == 'RNN':
        cemb = RNNEmbedding(len(label_encoder.char), cemb_dim,
                            padding_idx=label_encoder.char.get_pad(),
                            custom_lstm=custom_cemb_cell, dropout=dropout,
                            num_layers=cemb_layers, cell=cell, init_rnn=init_rnn)
    elif cemb_type.upper() == 'CNN':
        cemb = CNNEmbedding(len(label_encoder.char), cemb_dim,
                            padding_idx=label_encoder.char.get_pad())

    merger = None
    if cemb is not None and wemb is not None:
        if merge_type.lower() == 'mixer':
            if cemb.embedding_dim != wemb.embedding_dim:
                raise ValueError("EmbeddingMixer needs equal embedding dims")
            merger = EmbeddingMixer(wemb_dim)
            in_dim = wemb_dim
        elif merge_type == 'concat':
            merger = EmbeddingConcat()
            in_dim = wemb_dim + cemb.embedding_dim
        else:
            raise ValueError("Unknown merge method: {}".format(merge_type))
    elif cemb is not None:
        in_dim = cemb.embedding_dim
    else:
        in_dim = wemb_dim

    return (wemb, cemb, merger), in_dim


if __name__ == '__main__':
    from pie.settings import settings_from_file
    from pie.data import Dataset

    settings = settings_from_file('./config.json')
    data = Dataset(settings)
    ((word, wlen), (char, clen)), tasks = next(data.batch_generator())
    print("lemma", tasks['lemma'][0].size(), tasks['lemma'][1])
    print("char", char.size(), clen)
    print("word", word.size(), wlen)

    emb_dim = 20
    wemb = nn.Embedding(len(data.label_encoder.word), emb_dim)
    cemb = RNNEmbedding(len(data.label_encoder.char), emb_dim)
    cnncemb = CNNEmbedding(len(data.label_encoder.char), emb_dim)

    mixer = EmbeddingMixer(20)
    w, (c, _) = wemb(word), cemb(char, clen, wlen)
    output = mixer(w, c)

    output2 = []
    for w, c in zip(w, c):
        output2.append(mixer(w, c))
    output2 = torch.stack(output2)

    print(output.size())




import os
import tarfile
import datetime
import json
import time
import math
import random
import logging

import torch
import torch.nn.functional as F
import torch.nn as nn

import pie
from pie import utils, torch_utils, initialization
from pie.models import RNNEmbedding, CNNEmbedding, EmbeddingConcat, EmbeddingMixer
from pie.models import RNNEncoder, LinearDecoder


class EarlyStop(Exception):
    def __init__(self):
        pass


def load_sentences(path, bptt=35, buffer_size=100000):
    buf = []
    with open(path) as f:
        for line in f:
            line = line.strip()
            if not line:
                continue

            buf.extend(line.split())

            if len(buf) >= buffer_size:
                sents = [buf[i:i+bptt] for i in range(0, len(buf), bptt)]
                random.shuffle(sents)
                yield from sents
                buf = buf[len(sents)*bptt:]


def get_batches(lines, batch_size, label_encoder, device):
    batch = []
    for line in lines:
        if len(batch) == batch_size:
            (word, char), _ = label_encoder.transform(batch)
            w, wlen = torch_utils.pad_batch(
                word, label_encoder.word.get_pad(), device=device)
            c, clen = torch_utils.pad_batch(
                char, label_encoder.char.get_pad(), device=device)
            yield (w, wlen), (c, clen)
            batch = []
        batch.append(line)


class Encoder(nn.Module):
    def __init__(self, label_encoder, wemb_dim, cemb_dim, hidden_size, num_layers,
                 dropout=0.0, word_dropout=0.0, merge_type='concat', cemb_type='RNN',
                 cemb_layers=1, cell='LSTM', custom_cemb_cell=False,
                 init_rnn='xavier_uniform'):

        self.label_encoder = label_encoder
        self.wemb_dim = wemb_dim
        self.cemb_dim = cemb_dim
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        # kwargs
        self.cell = cell
        self.dropout = dropout
        self.word_dropout = word_dropout
        self.merge_type = merge_type
        self.cemb_type = cemb_type
        self.cemb_layers = cemb_layers
        self.custom_cemb_cell = custom_cemb_cell
        super().__init__()

        self.wemb = None
        if self.wemb_dim > 0:
            self.wemb = nn.Embedding(len(label_encoder.word), wemb_dim,
                                     padding_idx=label_encoder.word.get_pad())
            # init embeddings
            initialization.init_embeddings(self.wemb)

        self.cemb = None
        if cemb_type.upper() == 'RNN':
            self.cemb = RNNEmbedding(
                len(label_encoder.char), cemb_dim,
                padding_idx=label_encoder.char.get_pad(),
                custom_lstm=custom_cemb_cell, dropout=dropout,
                num_layers=cemb_layers, cell=cell, init_rnn=init_rnn)
        elif cemb_type.upper() == 'CNN':
            self.cemb = CNNEmbedding(
                len(label_encoder.char), cemb_dim,
                padding_idx=label_encoder.char.get_pad())

        self.merger = None
        if self.cemb is not None and self.wemb is not None:
            if merge_type.lower() == 'mixer':
                if self.cemb.embedding_dim != self.wemb.embedding_dim:
                    raise ValueError("EmbeddingMixer needs equal embedding dims")
                self.merger = EmbeddingMixer(wemb_dim)
                in_dim = wemb_dim
            elif merge_type == 'concat':
                self.merger = EmbeddingConcat()
                in_dim = wemb_dim + self.cemb.embedding_dim
            else:
                raise ValueError("Unknown merge method: {}".format(merge_type))
        elif self.cemb is None:
            in_dim = wemb_dim
        else:
            in_dim = self.cemb.embedding_dim

        # Encoder
        self.encoder = RNNEncoder(
            in_dim, hidden_size, num_layers=num_layers, cell=cell, dropout=dropout,
            init_rnn=init_rnn)

        # decoders
        self.lm_fwd_decoder = LinearDecoder(label_encoder.word, hidden_size)
        self.lm_bwd_decoder = LinearDecoder(label_encoder.word, hidden_size)

    def get_args_and_kwargs(self):
        return {'args': (self.wemb_dim, self.cemb_dim, self.hidden_size, self.num_layers),
                'kwargs': {'merge_type': self.merge_type, 'cemb_type': self.cemb_type,
                           'cemb_layers': self.cemb_layers, 'cell': self.cell,
                           'custom_cemb_cell': self.custom_cemb_cell}}

    def save(self, path):
        path = utils.ensure_ext(path, 'tar')
        # create dir if needed
        dirname = os.path.dirname(path)
        if dirname and not os.path.isdir(dirname):
            os.makedirs(dirname)

        with tarfile.open(path, 'w') as tar:
            # serialize label_encoder
            string, path = json.dumps(self.label_encoder.jsonify()), 'label_encoder.zip'
            utils.add_gzip_to_tar(string, path, tar)

            # serialize parameters
            string, path = json.dumps(self.get_args_and_kwargs()), 'parameters.zip'
            utils.add_gzip_to_tar(string, path, tar)

            # serialize weights
            with utils.tmpfile() as tmppath:
                torch.save(self.state_dict(), tmppath)
                tar.add(tmppath, arcname='state_dict.pt')

            # serialize current pie commit
            string, path = pie.__commit__, 'pie-commit.zip'
            utils.add_gzip_to_tar(string, path, tar)

    @classmethod
    def load(cls, path):
        with tarfile.open(utils.ensure_ext(path, 'tar'), 'r') as tar:
            commit = utils.get_gzip_from_tar(tar, 'pie-commit.zip')
            if pie.__commit__ != commit:
                logging.warn(
                    ("Model {} was serialized with a previous "
                     "version of `pie`. This might result in issues. "
                     "Model commit is {}, whereas current `pie` commit is {}."
                    ).format(path, commit, pie.__commit__))

            # load label encoder
            le = pie.dataset.MultiLabelEncoder.load_from_string(
                utils.get_gzip_from_tar(tar, 'label_encoder.zip'))

            # load model parameters
            params = json.loads(utils.get_gzip_from_tar(tar, 'parameters.zip'))

            # instantiate model
            model = Encoder(le, *params['args'], **params['kwargs'])

            # load state_dict
            with utils.tmpfile() as tmppath:
                tar.extract('state_dict.pt', path=tmppath)
                dictpath = os.path.join(tmppath, 'state_dict.pt')
                model.load_state_dict(torch.load(dictpath, map_location='cpu'))

        model.eval()

        return model

    def embedding(self, word, wlen, char, clen):
        wemb = cemb = None
        if self.wemb is not None:
            # set words to unknown with prob `p` depending on word frequency
            word = torch_utils.word_dropout(
                word, self.word_dropout, self.training, self.label_encoder.word)
            wemb = self.wemb(word)
        if self.cemb is not None:
            # cemb_outs: (seq_len x batch x emb_dim)
            cemb, _ = self.cemb(char, clen, wlen)

        if wemb is None:
            emb = cemb
        elif cemb is None:
            emb = wemb
        else:
            emb = self.merger(wemb, cemb)

        return emb

    def loss(self, word, wlen, char, clen):
        # Embedding
        emb = self.embedding(word, wlen, char, clen)

        # Encoder
        emb = F.dropout(emb, p=self.dropout, training=self.training)
        enc_outs = self.encoder(emb, wlen)

        fwd, bwd = F.dropout(enc_outs[-1], p=0, training=self.training).chunk(2, dim=2)
        # forward logits
        logits = self.lm_fwd_decoder(torch_utils.pad(fwd[:-1], pos='pre'))
        fwd_lm = self.lm_fwd_decoder.loss(logits, word)
        # backward logits
        logits = self.lm_bwd_decoder(torch_utils.pad(bwd[1:], pos='post'))
        bwd_lm = self.lm_bwd_decoder.loss(logits, word)

        return fwd_lm, bwd_lm

    def evaluate(self, dev):
        nbatches = 0
        tfwd_loss, tbwd_loss = 0, 0
        for (w, wlen), (c, clen) in dev:
            nbatches += 1
            fwd_loss, bwd_loss = self.loss(w, wlen, c, clen)
            tfwd_loss += fwd_loss.item()
            tbwd_loss += bwd_loss.item()

        return tfwd_loss / nbatches, tbwd_loss / nbatches


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--path', required=True)
    parser.add_argument('--devpath', required=True)
    parser.add_argument('--vocabpath')
    # model
    parser.add_argument('--max_size', type=int, default=50000)
    parser.add_argument('--wemb_dim', type=int, default=64)
    parser.add_argument('--cemb_dim', type=int, default=100)
    parser.add_argument('--cemb_type', default='rnn')
    parser.add_argument('--hidden_size', type=int, default=100)
    parser.add_argument('--num_layers', type=int, default=1)
    parser.add_argument('--cell', default='LSTM')
    # training
    parser.add_argument('--load_pretrained_embeddings')
    parser.add_argument('--dropout', type=float, default=0.0)
    parser.add_argument('--clipping', type=float, default=2.5)
    parser.add_argument('--patience', type=int, default=3)
    parser.add_argument('--word_dropout', type=float, default=0.0)
    parser.add_argument('--drop_hidden', type=float, default=0.0)
    parser.add_argument('--epochs', type=int, default=10)
    parser.add_argument('--batch_size', type=int, default=20)
    parser.add_argument('--bptt', type=int, default=35)
    parser.add_argument('--optim', default='SGD')
    parser.add_argument('--lr', type=float, default=0.1)
    parser.add_argument('--lr_weight', type=float, default=0.75)
    parser.add_argument('--buffer_size', type=int, default=1e+7)
    parser.add_argument('--weight_decay', type=float, default=2e-6)
    parser.add_argument('--device', default='cpu')
    parser.add_argument('--repfreq', type=int, default=100)
    parser.add_argument('--checkfreq', type=int, default=5000)
    parser.add_argument('--save', action='store_true')
    args = parser.parse_args()

    print("Fitting encoder")
    label_encoder = pie.dataset.MultiLabelEncoder(word_max_size=args.max_size)
    vocabpath = args.vocabpath or args.path
    label_encoder.fit(load_sentences(vocabpath))
    print(label_encoder)

    # model
    print("Building model")
    model = Encoder(label_encoder, args.wemb_dim, args.cemb_dim,
                    args.hidden_size, num_layers=args.num_layers, cell=args.cell,
                    cemb_type=args.cemb_type, dropout=args.dropout,
                    word_dropout=args.word_dropout)
    print(model)
    print(" * Number of parameters", sum(p.nelement() for p in model.parameters()))

    if args.load_pretrained_embeddings:
        print("Loading pretrained embeddings from", args.load_pretrained_embeddings)
        pie.initialization.init_pretrained_embeddings(
            args.load_pretrained_embeddings, label_encoder.word, model.wemb)

    model.to(args.device)
    print("Starting training")
    print()

        # optim
    if args.optim.lower() == 'sgd':
        optim = torch.optim.SGD(list(model.parameters()), lr=args.lr, # momentum=0.9,
                                weight_decay=args.weight_decay)
    else:
        optim = getattr(torch.optim, args.optim)(list(model.parameters()), lr=args.lr,
                                                 weight_decay=args.weight_decay)

    # report
    best_loss, best_params, fails = float('inf'), None, 0

    try:
        for epoch in range(args.epochs):
            print("Starting epoch:", epoch + 1)
            tbatches, tfwd_loss, tbwd_loss, titems, ttime = 0, 0, 0, 0, time.time()
            rbatches, rfwd_loss, rbwd_loss, ritems, rtime = 0, 0, 0, 0, time.time()

            for (w, wlen), (c, clen) in get_batches(
                    load_sentences(args.path), args.batch_size,
                    label_encoder, args.device):
    
                fwd_loss, bwd_loss = model.loss(w, wlen, c, clen)
    
                # optimize
                optim.zero_grad()
                nn.utils.clip_grad_norm_(model.parameters(), args.clipping)
                ((fwd_loss + bwd_loss) / 2).backward()
                optim.step()
    
                # accumulate
                tfwd_loss += fwd_loss.item()
                tbwd_loss += bwd_loss.item()
                titems += sum(wlen)
                tbatches += 1
                rfwd_loss += fwd_loss.item()
                rbwd_loss += bwd_loss.item()
                ritems += sum(wlen)
                rbatches += 1
                
                if rbatches == args.repfreq:
                    print('{}/{}: fwd loss=>{:.3f} bwd loss=>{:.3f} '.format(
                        epoch + 1, tbatches,
                        math.exp(rfwd_loss/rbatches),
                        math.exp(rbwd_loss/rbatches)) +
                          'speed=>{:.3f} w/s'.format(ritems / (time.time()-rtime)))
                    rbatches, rfwd_loss, rbwd_loss, ritems = 0, 0, 0, 0
                    rtime = time.time()
    
                if tbatches % args.checkfreq == 0:
                    # do validation
                    model.eval()
                    with torch.no_grad():
                        dev_fwd_loss, dev_bwd_loss = model.evaluate(
                            get_batches(load_sentences(args.devpath),
                                        args.batch_size, label_encoder, args.device))
                    model.train()
            
                    print("Evaluation: fwd=>{:.3f}, bwd=>{:.3f}".format(
                        math.exp(dev_fwd_loss), math.exp(dev_bwd_loss)))
            
                    if (dev_fwd_loss + dev_bwd_loss) / 2 < best_loss:
                        best_loss = (dev_fwd_loss + dev_bwd_loss) / 2
                        best_params = model.to('cpu').state_dict()
                        model.to(args.device)
                        fails = 0
                    else:
                        fails += 1
                        # reduce LR
                        for pgroup in optim.param_groups:
                            pgroup['lr'] = pgroup['lr'] * args.lr_weight
                        print(optim)
            
                    if fails == args.patience:
                        raise EarlyStop

            # epoch
            print("Epoch: {}; loss: fwd=>{:.3f}, bwd=>{:.3f}; speed=>{:.3f} w/sec".format(
                epoch + 1,
                math.exp(tfwd_loss / tbatches),
                math.exp(tbwd_loss / tbatches),
                titems / (time.time()-ttime)))

    except EarlyStop:
        print("Stop training with best loss=>{:.3f}".format(math.exp(best_loss)))

    if args.save:
        model.to('cpu').load_state_dict(best_params)
        timestamp = datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
        model.save(os.path.join('pretrained-encoder', timestamp))
